from __future__ import annotations

import logging
import os
from collections.abc import Generator, Sequence
from typing import TYPE_CHECKING, Any, TextIO, TypeVar

from . import fields
from .site import Path, SiteElement

if TYPE_CHECKING:
    from .page import Page
    from .site import Site

log = logging.getLogger("node")

P = TypeVar("P", bound="Page")


class SkipPage(Exception):
    """
    Exception raised when a page should not be added to the site
    """


class Node(SiteElement):
    """
    One node in the rendered directory hierarchy of the site.

    Each node corresponds to a directory in the built site.

    Most nodes also correspond to directories in the source site, with the
    exceptions of nodes for autogenerated pages, and nodes for pages that will
    appear as paths (e.g.: `foo.md` builds as `foo/index.html`).
    """

    site_path = fields.Str["Node"](
        doc="""
        Where a content directory appears in the site.

        By default, is is the `site_path` of the parent directory, plus the directory
        name.

        If you are publishing the site at `/prefix` instead of the root of the domain,
        override this with `/prefix` in the content root.
    """
    )

    def __init__(self, site: Site, name: str, *, parent: Node | None = None):
        super().__init__(site, parent=parent)
        # Basename of this node
        self.name: str = name
        # Path of this node from the root of the site (without leading /)
        self.path: str
        if parent is None:
            self.path = name
        else:
            self.path = os.path.join(parent.path, name)
        # Parent node, or None if this is the root
        self.parent: Node | None = parent
        # Index page for this directory, if present
        self.page: Page | None = None
        # SourcePages in this node indexed by source_name
        self.by_src_relpath: dict[str, Page] = {}
        # Pages to be rendered at this location
        self.build_pages: dict[str, Page] = {}
        # Subdirectories
        self.sub: dict[str, Node] = {}

    def __repr__(self) -> str:
        return f"Node({self.name})"

    def is_empty(self) -> bool:
        """
        Check if this node does not contain any content
        """
        if self.page is not None:
            return False
        if self.by_src_relpath:
            return False
        if self.build_pages:
            return False
        if self.sub:
            return False
        return True

    def prune_empty_subnodes(self) -> None:
        """
        Prune empty subnodes
        """
        empty: list[str] = []

        for name, node in self.sub.items():
            if node.is_empty():
                empty.append(name)

        # Remve empty subnodes
        for name in empty:
            del self.sub[name]

    def print(self, lead: str = "", file: TextIO | None = None) -> None:
        if self.page:
            print(f"{lead}{self.name!r} page:{self.page!r}", file=file)
        else:
            print(f"{lead}{self.name!r} (no page)", file=file)
        for name, page in self.build_pages.items():
            print(f"{lead}- {name} → {page!r}", file=file)
        for name, node in self.sub.items():
            node.print(lead + "+ ", file=file)

    def iter_pages(
        self, prune: Sequence[Node] = (), source_only: bool = False
    ) -> Generator[Page, None, None]:
        """
        Iterate all pages in this subtree
        """
        # Avoid nodes in prune
        if self in prune:
            return
        if source_only:
            yield from self.by_src_relpath.values()
        else:
            yield from self.build_pages.values()
        for node in self.sub.values():
            yield from node.iter_pages(prune, source_only=source_only)

    def lookup_page(self, path: Path) -> Page | None:
        # print(f"Node.lookup_page: {self.name=!r}, {self.page=!r}, {path=!r},"
        #       f" sub={self.sub.keys() if self.sub else 'None'}")

        if not path:
            return self.page

        if path.head == "..":
            if self.parent is None:
                return None
            return self.parent.lookup_page(path.tail)
        elif path.head in (".", ""):
            # Probably not worth trying to avoid a recursion step here, since
            # this should not be a common occurrence
            return self.lookup_page(path.tail)

        if len(path) > 1:
            if not self.sub:
                return None
            elif subnode := self.sub.get(path.head):
                # print(f"Node.lookup_page:  descend into {subnode.name!r} with {path.tail=!r}")
                return subnode.lookup_page(path.tail)
            else:
                return None

        # Match subnode name
        if (subnode := self.sub.get(path.head)) and subnode.page:
            return subnode.page

        # Match subpage names and basename of src.relpath in subpages
        if (page := self.build_pages.get(path.head)) is not None:
            return page
        if (page := self.by_src_relpath.get(path.head)) is not None:
            return page

        return None

    def create_auto_page_as_file(
        self, *, page_cls: type[P], dst: str, **kw: Any
    ) -> P | None:
        """
        Create a page of the given type, attaching it at the given path
        """
        if "src" in kw:
            raise RuntimeError("auto page created with 'src' set")

        if self.site.last_load_step != self.site.LOAD_STEP_ORGANIZE:
            raise RuntimeError(
                "Node.create_auto_page created outside the 'generate' step"
            )

        try:
            if dst:
                return self._create_leaf_page(page_cls=page_cls, dst=dst, **kw)
            else:
                return self._create_index_page(
                    page_cls=page_cls, directory_index=False, **kw
                )
        except SkipPage:
            return None

    def create_auto_page_as_path(
        self, page_cls: type[P], name: str, **kw: Any
    ) -> P | None:
        """
        Create a page of the given type, attaching it at the given path
        """
        if "src" in kw:
            raise RuntimeError("auto page created with 'src' set")

        if self.site.last_load_step != self.site.LOAD_STEP_ORGANIZE:
            raise RuntimeError(
                "Node.create_auto_page created outside the 'generate' step"
            )

        node = self._child(name)

        try:
            return node._create_index_page(
                page_cls=page_cls, directory_index=False, **kw
            )
        except SkipPage:
            return None

    def create_auto_page_as_index(self, page_cls: type[P], **kw: Any) -> P | None:
        """
        Create a page of the given type, attaching it at the given path
        """
        if "src" in kw:
            raise RuntimeError("auto page created with 'src' set")

        if self.site.last_load_step != self.site.LOAD_STEP_ORGANIZE:
            raise RuntimeError(
                "Node.create_auto_page created outside the 'generate' step"
            )

        try:
            return self._create_index_page(
                page_cls=page_cls, directory_index=True, **kw
            )
        except SkipPage:
            return None

    def _create_index_page(
        self, *, page_cls: type[P], directory_index: bool = False, **kw: Any
    ) -> P:
        from .asset import Asset
        from .page import PageValidationError

        if directory_index:
            search_root_node = self
        elif self.parent is not None:
            search_root_node = self.parent
        else:
            search_root_node = self

        # Create the page
        try:
            page = self.site.features.get_page_class(page_cls)(
                site=self.site,
                dst="index.html",
                node=self,
                search_root_node=search_root_node,
                leaf=False,
                directory_index=directory_index,
                **kw,
            )
        except PageValidationError as e:
            log.warning("%s: skipping page: %s", e.page, e.msg)
            raise SkipPage()

        if self.page is not None:
            if isinstance(self.page, Asset) and isinstance(page, Asset):
                # Allow replacement of assets
                page.old_footprint = self.page.old_footprint
            else:
                log.warning(
                    "%s: page %r attempts to replace page %r: skipped",
                    self.path,
                    page,
                    self.page,
                )
                raise SkipPage()

        self.page = page
        if page.source_name is not None:
            search_root_node.by_src_relpath[page.source_name] = page
        self.build_pages["index.html"] = page
        # if page.directory_index is False:
        #     print(f"{page=!r} dst is not set but page is not a directory index")
        return page

    def _create_leaf_page(self, *, page_cls: type[P], dst: str, **kw: Any) -> P:
        from .asset import Asset
        from .page import PageValidationError

        # Create the page
        try:
            page = self.site.features.get_page_class(page_cls)(
                site=self.site,
                dst=dst,
                node=self,
                search_root_node=self,
                leaf=True,
                directory_index=False,
                **kw,
            )
        except PageValidationError as e:
            log.warning("%s: skipping page: %s", e.page, e.msg)
            raise SkipPage()

        if old := self.build_pages.get(dst):
            if isinstance(old, Asset) and isinstance(page, Asset):
                # Allow replacement of assets
                page.old_footprint = old.old_footprint
            else:
                log.warning(
                    "%s: page %r attempts to replace page %r: skipped",
                    self.path,
                    page,
                    self.page,
                )
                raise SkipPage()

        self.build_pages[dst] = page
        if page.source_name is not None:
            self.by_src_relpath[page.source_name] = page
        return page

    def _child(self, name: str) -> Node:
        """
        Return the given subnode, creating it if missing
        """
        if node := self.sub.get(name):
            return node

        node = self.site.features.get_node_class(Node)(
            site=self.site, name=name, parent=self
        )
        self.sub[name] = node
        return node

    def lookup_node(self, path: Path) -> Node | None:
        """
        Return the subnode at the given relative path, or None if it does not
        exist.

        Path elements of "." and ".." are supported
        """
        if not path:
            return self
        if path.head == ".":
            # Probably not worth trying to avoid a recursion step here, since
            # this should not be a common occurrence
            return self.lookup_node(path.tail)
        elif path.head == "..":
            if self.parent is None:
                return None
            else:
                return self.parent.lookup_node(path.tail)
        elif sub := self.sub.get(path.head):
            return sub.lookup_node(path.tail)
        return None

    def contains(self, page: Page) -> bool:
        """
        Check if page is contained in or under this node
        """
        # Walk the parent chain of page.node to see if we find self
        node: Node | None = page.node
        while node is not None:
            if node == self:
                return True
            node = node.parent
        return False



from __future__ import annotations

import argparse
import logging
import sys
from typing import Any

from staticsite.page_filter import compile_page_match
from staticsite.utils import front_matter

from .command import SiteCommand, register

log = logging.getLogger("dump_meta")


@register
class DumpMeta(SiteCommand):
    "Dump the metadata of each page in the site"

    NAME = "dump_meta"

    @classmethod
    def add_subparser(
        cls, subparsers: argparse._SubParsersAction[Any]
    ) -> argparse.ArgumentParser:
        parser = super().add_subparser(subparsers)
        parser.add_argument(
            "-f",
            "--format",
            action="store",
            default="yaml",
            help="format to use for output",
        )
        parser.add_argument(
            "-p",
            "--pages",
            nargs="+",
            default=(),
            help="globs or regexps matching pages to show",
        )
        return parser

    def run(self) -> None:
        site = self.load_site()
        filters = [compile_page_match(f) for f in self.args.pages]
        res = {}
        # show_repr = self.args.repr
        for site_path, page in sorted((p.site_path, p) for p in site.iter_pages()):
            if filters and not any(f.match(site_path) for f in filters):
                continue
            res[site_path] = page.to_dict()
        sys.stdout.write(front_matter.write(res, self.args.format))



from __future__ import annotations

import os
from unittest import TestCase

from . import utils as test_utils


class TestBlog(test_utils.SiteTestMixin, TestCase):
    site_name = "blog"
    site_settings = {"SITE_AUTHOR": "Test User"}

    @test_utils.assert_no_logs()
    def test_render_paths(self):
        self.assertBuilt(
            "index.md",
            "",
            "index.html",
            sample="Hello. These are the last 10 posts of my blog",
        )
        self.assertBuilt(
            "about.md",
            "about",
            "about/index.html",
            sample="it can be a starting point for more",
        )
        self.assertBuilt(
            "posts/example.md",
            "posts/example",
            "posts/example/index.html",
            sample="This is an example blog post",
        )

    def test_meta(self):
        self.maxDiff = None

        self.assertPagePaths(
            [
                "",
                "index.rss",
                "index.atom",
                "archive",
                "about",
                "posts",
                "posts/example",
                "posts/example.jpg",
                "posts/example-small.jpg",
                "posts/example-thumbnail.jpg",
            ]
        )

        index, page, archive, image = self.mocksite.page(
            "", "about", "archive", "posts/example.jpg"
        )
        self.assertEqual(page.meta["title"], "About")

        self.assertEqual(
            image.get_img_attributes(),
            {
                "alt": "This is an example image",
                "src": "/posts/example.jpg",
                "srcset": "/posts/example-small.jpg 480w, /posts/example-thumbnail.jpg 128w, /posts/example.jpg 500w",
            },
        )

        self.assertEqual(
            page.to_dict(),
            {
                "src": {
                    "relpath": "about.md",
                    "abspath": os.path.join(self.site.content_root, "about.md"),
                },
                "site_path": "about",
                "build_path": "about/index.html",
                "meta": {
                    "date": "2019-06-01 12:30:00+02:00",
                    "draft": False,
                    "author": "Test User",
                    "copyright": "© 2019 Test User",
                    "indexed": True,
                    "syndicated": True,
                    "syndication_date": "2019-06-01 12:30:00+02:00",
                    "site_name": "My example blog",
                    "site_url": "https://www.example.org",
                    "template": "page.html",
                    "template_copyright": "compiled:None",
                    "title": "About",
                    "nav": ["MarkdownPage(about)"],
                    "nav_title": "About",
                    "related": {},
                },
                "type": "markdown",
            },
        )

        self.assertEqual(
            index.to_dict(),
            {
                "src": {
                    "relpath": "index.md",
                    "abspath": os.path.join(self.site.content_root, "index.md"),
                },
                "site_path": "",
                "build_path": "index.html",
                "meta": {
                    "date": "2019-12-30 17:30:00+01:00",
                    "draft": False,
                    "author": "Test User",
                    "copyright": "© 2019 Test User",
                    "indexed": True,
                    "syndicated": True,
                    "syndication": {
                        "add_to": True,
                        "archive": {
                            "index": "MarkdownPage()",
                            "pages": ["MarkdownPage(posts/example)"],
                        },
                        "archive_page": "ArchivePage(archive)",
                        "atom_page": "AtomPage(index.atom)",
                        "index": "MarkdownPage()",
                        "pages": ["MarkdownPage(posts/example)"],
                        "rss_page": "RSSPage(index.rss)",
                    },
                    "syndication_date": "2019-12-30 17:30:00+01:00",
                    "pages": ["MarkdownPage(posts/example)"],
                    "site_name": "My example blog",
                    "site_url": "https://www.example.org",
                    "template": "blog.html",
                    "template_copyright": "compiled:None",
                    "title": "My example blog",
                    "nav": ["MarkdownPage(about)"],
                    "related": {
                        "archive": "ArchivePage(archive)",
                        "atom_feed": "AtomPage(index.atom)",
                        "rss_feed": "RSSPage(index.rss)",
                    },
                },
                "type": "markdown",
            },
        )
        # A blog page renders images, relative links, only the beginning of
        # split pages
        rendered = index.render().content()
        self.assertNotIn(b"This is the rest of the blog post", rendered)
        self.assertIn(b"srcset='/posts/example", rendered)

        self.assertEqual(
            archive.to_dict(),
            {
                "site_path": "archive",
                "build_path": "archive/index.html",
                "meta": {
                    "date": "2019-12-30 17:30:00+01:00",
                    "syndication_date": "2019-12-30 17:30:00+01:00",
                    "author": "Test User",
                    "copyright": "© 2019 Test User",
                    "created_from": "MarkdownPage()",
                    "index": "MarkdownPage()",
                    "indexed": False,
                    "syndicated": False,
                    "pages": ["MarkdownPage(posts/example)"],
                    "site_name": "My example blog",
                    "site_url": "https://www.example.org",
                    "template": "archive.html",
                    "template_copyright": "compiled:None",
                    "title": "My example blog",
                    "nav": ["MarkdownPage(about)"],
                    "related": {
                        "atom_feed": "AtomPage(index.atom)",
                        "rss_feed": "RSSPage(index.rss)",
                    },
                },
                "type": "archive",
            },
        )

        self.assertEqual(
            self.page("index.rss").to_dict(),
            {
                "site_path": "index.rss",
                "build_path": "index.rss",
                "meta": {
                    "created_from": "MarkdownPage()",
                    "date": "2019-12-30 17:30:00+01:00",
                    "syndication_date": "2019-12-30 17:30:00+01:00",
                    "index": "MarkdownPage()",
                    "author": "Test User",
                    "copyright": "© 2019 Test User",
                    "indexed": False,
                    "syndicated": False,
                    "pages": ["MarkdownPage(posts/example)"],
                    "site_name": "My example blog",
                    "site_url": "https://www.example.org",
                    "template": "syndication.rss",
                    "template_copyright": "compiled:None",
                    "title": "My example blog",
                    "nav": ["MarkdownPage(about)"],
                    "related": {},
                },
                "type": "rss",
            },
        )
        # A feed page renders images, all of a split page, absolute site urls
        rendered = self.page("index.rss").render().content()
        self.assertIn(
            b"src=&#39;https://www.example.org/posts/example-small.jpg", rendered
        )
        self.assertIn(b"This is the rest of the blog post", rendered)

        self.assertEqual(
            image.to_dict(),
            {
                "src": {
                    "relpath": "posts/example.jpg",
                    "abspath": os.path.join(
                        self.site.content_root, "posts/example.jpg"
                    ),
                },
                "site_path": "posts/example.jpg",
                "build_path": "posts/example.jpg",
                "meta": {
                    "date": "2019-06-01 12:30:00+02:00",
                    "syndication_date": "2019-06-01 12:30:00+02:00",
                    "draft": False,
                    "author": "Test User",
                    "copyright": "© 2019 Test User",
                    "indexed": False,
                    "syndicated": False,
                    "site_name": "My example blog",
                    "site_url": "https://www.example.org",
                    "title": "This is an example image",
                    "template_copyright": "compiled:None",
                    "width": 500,
                    "height": 477,
                    "nav": ["MarkdownPage(about)"],
                    "related": {
                        "small": "ScaledImage(posts/example-small.jpg)",
                        "thumbnail": "ScaledImage(posts/example-thumbnail.jpg)",
                    },
                },
                "type": "image",
            },
        )

        self.assertEqual(
            self.page("posts/example-small.jpg").to_dict(),
            {
                "site_path": "posts/example-small.jpg",
                "build_path": "posts/example-small.jpg",
                "meta": {
                    "created_from": "Image(posts/example.jpg)",
                    "date": "2019-06-01 12:30:00+02:00",
                    "syndication_date": "2019-06-01 12:30:00+02:00",
                    "author": "Test User",
                    "copyright": "© 2019 Test User",
                    "indexed": False,
                    "syndicated": False,
                    "site_name": "My example blog",
                    "site_url": "https://www.example.org",
                    "template_copyright": "compiled:None",
                    "title": "This is an example image",
                    "width": 480,
                    "height": 458,
                    "nav": ["MarkdownPage(about)"],
                    "related": {},
                },
                "type": "scaledimage",
            },
        )

        self.assertEqual(
            self.page("posts/example-thumbnail.jpg").to_dict(),
            {
                "site_path": "posts/example-thumbnail.jpg",
                "build_path": "posts/example-thumbnail.jpg",
                "meta": {
                    "created_from": "Image(posts/example.jpg)",
                    "date": "2019-06-01 12:30:00+02:00",
                    "syndication_date": "2019-06-01 12:30:00+02:00",
                    "author": "Test User",
                    "copyright": "© 2019 Test User",
                    "indexed": False,
                    "syndicated": False,
                    "site_name": "My example blog",
                    "site_url": "https://www.example.org",
                    "template_copyright": "compiled:None",
                    "title": "This is an example image",
                    "width": 128,
                    "height": 122,
                    "nav": ["MarkdownPage(about)"],
                    "related": {},
                },
                "type": "scaledimage",
            },
        )



from __future__ import annotations

import locale
import logging
import mimetypes
import os
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from staticsite.node import Node
    from staticsite.page import Page
    from staticsite.site import Site

log = logging.getLogger("serve")


class PageFS:
    """
    VFS-like abstraction that maps the names of files that pages would render
    with the corresponding pages.

    This can be used to render pages on demand.
    """

    def __init__(self, site: Site):
        self.paths: dict[str, Page] = {}
        self.add_tree(site.root)

    def add_tree(self, root: Node, relpath: str = "") -> None:
        for name, page in root.build_pages.items():
            self.paths[os.path.join(relpath, name)] = page

        for name, sub in root.sub.items():
            self.add_tree(sub, os.path.join(relpath, name))

    def get_page(self, relpath: str) -> tuple[None, None] | tuple[str, Page]:
        if not relpath:
            relpath = "/index.html"
        dst_relpath = os.path.normpath(relpath).lstrip("/")
        page = self.paths.get(dst_relpath, None)
        if page is not None:
            return dst_relpath, page

        dst_relpath = os.path.join(dst_relpath, "index.html")
        page = self.paths.get(dst_relpath, None)
        if page is not None:
            return dst_relpath, page

        return None, None

    def render(self, path: str, **kw: Any) -> tuple[str | None, bytes | None]:
        """
        Find the page for the given path and return its rendered contents
        """
        dst_relpath, page = self.get_page(os.path.normpath(path).lstrip("/"))
        if page is None:
            return None, None

        # Set locale for rendering
        try:
            lname = page.site.settings.LANGUAGES[0]["locale"]
            locale.setlocale(locale.LC_ALL, lname)
        except locale.Error as e:
            log.warning("%s: cannot set locale to %s: %s", page, lname, e)

        rendered = page.render(**kw)
        build_path = os.path.join(page.node.path, page.dst)
        return build_path, rendered.content()

    # FIXME: using a generic callable until I find the right type for the WSGI start_response interface
    def serve_path(
        self, path: str, environ: dict[str, Any], start_response: Callable[..., Any]
    ) -> list[bytes]:
        """
        Render a page on the fly and serve it.

        Call start_response with the page headers and return the bytes() with
        the page contents.

        start_response is the start_response from WSGI

        Returns None without calling start_response if no page was found.
        """
        relpath, content = self.render(path)
        if relpath is None or content is None:
            start_response(
                "404 NOT FOUND",
                [
                    ("Content-Type", "text/plain"),
                    ("Content-Length", "0"),
                ],
            )
            return []
        else:
            start_response(
                "200 OK",
                [
                    ("Content-Type", mimetypes.guess_type(relpath)[0]),
                    ("Content-Length", str(len(content))),
                ],
            )
            return [content]



from __future__ import annotations

import os
import sys
from collections.abc import Sequence
from typing import Any

# Default settings

# Root directory used to resolve relative path in settings
# Default if None: the directory where the settings file is found
PROJECT_ROOT: str | None = None

# Base URL for the site, used to generate absolute URLs
SITE_URL: str | None = None

# Root directory of the site in the URLs we generate.
#
# If you are publishing the site at /prefix instead of root of the domain,
# override this with /prefix
SITE_ROOT: str = "/"

# Default site name. If None, use the title of the toplevel index
SITE_NAME: str | None = None

# Default author of the site
SITE_AUTHOR: str | None = None

# Directory with "archetypes" (templates used by ssite new)
# If None, archetypes are not used by ssite new
ARCHETYPES: str = "archetypes"

# Directory with the source content of the site
# Default if None: PROJECT_ROOT
CONTENT: str | None = None

# Directories where themes are looked for
THEME_PATHS: Sequence[str] = [
    os.path.abspath(
        os.path.join(os.path.dirname(os.path.realpath(sys.argv[0])), "themes")
    ),
    "/usr/share/staticsite/themes",
]

# Theme used to render the site.
# For compatibility, if it is a sequence of strings, it is treated as a list of
# full paths to theme directories to try in order
THEME: str | Sequence[str] = "default"

# Directory where the static site will be written by build
# If None, require providing it explicitly to build
OUTPUT: str | None = None

# Time zone used for timestamps on the site
# (NONE defaults to the system configured timezone)
TIMEZONE: str | None = None

# Editor used to edit new pages
EDITOR: str = os.environ.get("EDITOR", "sensible-editor")

# Command used to run the editor, as passed to subprocess.check_command.
# Each list element is expanded with string.format. All settings are available
# for expansion, and {name} is the absolute path of the file to edit.
EDIT_COMMAND: Sequence[str] = ["{EDITOR}", "{name}", "+"]

# extensions for python-markdown and their config used for this site
MARKDOWN_EXTENSIONS = [
    "markdown.extensions.extra",
    "markdown.extensions.codehilite",
    "markdown.extensions.fenced_code",
]
MARKDOWN_EXTENSION_CONFIGS = {
    "markdown.extensions.extra": {
        "markdown.extensions.footnotes": {
            # See https://github.com/spanezz/staticsite/issues/13
            "UNIQUE_IDS": True,
        },
    },
}

# List of asset directories included from /usr/share/javascript
SYSTEM_ASSETS: Sequence[str] = []

# If true, do not ignore pages with dates in the future
DRAFT_MODE: bool = False

# If True, store cached data to speed up rebuilds
CACHE_REBUILDS: bool = True

# Patterns (glob or regexps) that identify files in content directories that
# are parsed as jinja2 templates
JINJA2_PAGES: Sequence[str] = ["*.html", "*.j2.*"]

# Set to false to disable running Jinja2 in a sandboxed environnment.
# If you trust your site sources, it renders noticeably faster.
JINJA2_SANDBOXED = True

# Languages used to build the site
# For now, only the first one is used, and only its locale is used.
LANGUAGES: Sequence[dict[str, Any]] = [
    {
        "locale": "C",
    },
]

# Path where theme static assets will be placed in built site
# Override with "" to merge when with the rest of the contents
STATIC_PATH = "static"



from __future__ import annotations

import re
from unittest import TestCase

from . import utils as test_utils

SVG = """
<?xml version="1.0" standalone="no"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 20010904//EN"
  "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd">
<svg width="24" height="24">
  <text x="0" y="5">2</text>
</svg>
"""

XPM = """/* XPM */
static char * bottom_active_xpm[] = {
"24 3 3 1",
"       c None",
"#      c #C0C0C0 s active_color_2",
"@      c #C0C0FF s active_color_2",
"@@@@@@@@@@@@@@@@@@@@@@@@",
"@@@@@@@@@@@@@@@@@@@@@@@@",
"@@@@@@@@@@@@@@@@@@@@@@@@"};
"""


class TestImages(test_utils.MockSiteTestMixin, TestCase):
    def test_images(self):
        self.maxDiff = None
        files = {
            "index.md": {
                "syndication": True,
                "pages": "blog/*",
                "template": "blog.html",
                "title": "Blog",
            },
            "blog/post.md": """---
date: 2019-01-01 12:00:00+02:00
---

# Title

![Photo](images/photo.xpm)

![Photo](images/photo.svg)
""",
            "blog/images/photo.xpm": XPM,
            "blog/images/photo.svg": SVG,
        }
        with self.site(files) as mocksite:
            mocksite.assertPagePaths(
                (
                    "",
                    "index.rss",
                    "index.atom",
                    "archive",
                    "blog",
                    "blog/post",
                    "blog/images",
                    "blog/images/photo.xpm",
                    "blog/images/photo.svg",
                )
            )

            post = mocksite.page("")
            rendered = post.render().buf
            mo = re.search(r'src="([a-z/:.]+)/photo.xpm"', rendered.decode())
            self.assertTrue(mo)
            self.assertEqual(mo.group(1), "https://www.example.org/blog/images")
            mo = re.search(r'src="([a-z/:.]+)/photo.svg"', rendered.decode())
            self.assertTrue(mo)
            self.assertEqual(mo.group(1), "https://www.example.org/blog/images")

            post = mocksite.page("blog/post")
            rendered = post.render().buf
            mo = re.search(r'src="([a-z/:.]+)/photo.xpm"', rendered.decode())
            self.assertTrue(mo)
            self.assertEqual(mo.group(1), "/blog/images")
            mo = re.search(r'src="([a-z/:.]+)/photo.svg"', rendered.decode())
            self.assertTrue(mo)
            self.assertEqual(mo.group(1), "/blog/images")

            rss = mocksite.page("index.rss")
            rendered = rss.render().buf
            mo = re.search(r"src=&#34;([a-z/:.]+)/photo.xpm&#34;", rendered.decode())
            self.assertEqual(mo.group(1), "https://www.example.org/blog/images")
            mo = re.search(r"src=&#34;([a-z/:.]+)/photo.svg&#34;", rendered.decode())
            self.assertEqual(mo.group(1), "https://www.example.org/blog/images")



from __future__ import annotations

import os
import shutil
from typing import IO, Literal, overload

from .file import File


class RenderedElement:
    """
    Abstract interface for a rendered site page
    """

    def write(self, *, name: str, dir_fd: int, old: os.stat_result | None) -> None:
        """
        Write the rendered contents to the given file
        """
        raise NotImplementedError("{}.write", self.__class__.__name__)

    def content(self) -> bytes:
        """
        Return the rendered contents as bytes
        """
        raise NotImplementedError("{}.write", self.__class__.__name__)

    @overload
    @classmethod
    def dirfd_open(cls, name: str, mode: Literal["wt"], dir_fd: int) -> IO[str]:
        ...

    @overload
    @classmethod
    def dirfd_open(cls, name: str, mode: Literal["wb"], dir_fd: int) -> IO[bytes]:
        ...

    @classmethod
    def dirfd_open(
        cls, name: str, mode: Literal["wt", "wb"], dir_fd: int
    ) -> IO[str] | IO[bytes]:
        """
        Open a file contained in the directory pointed to by dir_fd
        """

        def _file_opener(fname: str, flags: int) -> int:
            return os.open(fname, flags, mode=0o666, dir_fd=dir_fd)

        return open(name, mode=mode, opener=_file_opener)


class RenderedFile(RenderedElement):
    def __init__(self, src: File):
        self.src = src

    def write(self, *, name: str, dir_fd: int, old: os.stat_result | None) -> None:
        try:
            st = os.stat(name, dir_fd=dir_fd)
        except FileNotFoundError:
            st = None

        if st is None or (
            self.src.stat.st_mtime > st.st_mtime or self.src.stat.st_size != st.st_size
        ):
            with open(self.src.abspath, "rb") as fd:
                with self.dirfd_open(name, "wb", dir_fd=dir_fd) as out:
                    shutil.copyfileobj(fd, out)
                    # TODO: copystat can accept unix file descriptors as
                    # arguments, but it is not typed accordingly
                    shutil.copystat(fd.fileno(), out.fileno())  # type: ignore

    def content(self) -> bytes:
        with open(self.src.abspath, "rb") as fd:
            return fd.read()


class RenderedString(RenderedElement):
    def __init__(self, s: str | None):
        if s is None:
            self.buf = b"Error in page: see build logs"
        else:
            self.buf = s.encode("utf-8")

    def write(self, *, name: str, dir_fd: int, old: os.stat_result | None) -> None:
        with self.dirfd_open(name, "wb", dir_fd=dir_fd) as out:
            out.write(self.buf)

    def content(self) -> bytes:
        return self.buf



from __future__ import annotations

import logging
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, cast

from staticsite import fields
from staticsite.feature import Feature, PageTrackingMixin, TrackedField
from staticsite.page import AutoPage, ChangeExtent, Page, SourcePage, TemplatePage
from staticsite.site import Path

if TYPE_CHECKING:
    from staticsite.features import rst

log = logging.getLogger("aliases")


class AliasField(TrackedField[Page, list[str]]):
    """
    Relative paths in the destination directory where the page should also show up.
    [Like in Hugo](https://gohugo.io/extras/aliases/), this can be used to maintain
    existing links when moving a page to a different location.
    """

    tracked_by = "aliases"

    def _clean(self, page: Page, value: Any) -> list[str]:
        if isinstance(value, str):
            return [value]
        elif isinstance(value, list):
            return value
        elif isinstance(value, tuple):
            return list(value)
        else:
            raise TypeError(
                f"invalid value of type {type(value)} for {page!r}.{self.name}:"
                " expecting, str, list[str], or tuple[str]"
            )


class SourcePageField(fields.Field["AliasPage", SourcePage]):
    """
    Field containing a SourcePage
    """

    def _clean(self, page: AliasPage, value: Any) -> SourcePage:
        if isinstance(value, SourcePage):
            return value
        else:
            raise TypeError(
                f"invalid value of type {type(value)} for {page!r}.{self.name}:"
                " expecting, SourcePage"
            )


class AliasesPageMixin(Page):
    aliases = AliasField(structure=True)


class AliasesFeature(PageTrackingMixin[AliasesPageMixin], Feature):
    """
    Build redirection pages for page aliases.

    A page can define 'aliases=[...]' to generate pages in those locations that
    redirect to the page.
    """

    RUN_AFTER = ["rst"]

    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)
        cast("rst.RestructuredText", self.site.features["rst"]).yaml_tags.add("aliases")

    def get_page_bases(self, page_cls: type[Page]) -> Sequence[type[Page]]:
        return (AliasesPageMixin,)

    def get_used_page_types(self) -> list[type[Page]]:
        return [AliasPage]

    def generate(self) -> None:
        # Build alias pages from pages with an 'aliases' metadata
        for page in self.tracked_pages:
            if not (aliases := page.aliases):
                continue

            for alias in aliases:
                path = Path.from_string(alias)
                node = self.site.root.generate_path(path.dir)
                if old := node.lookup_page(Path((path.name,))):
                    if old == page:
                        log.warning(
                            "%r defines alias %r pointing to itself", page, alias
                        )
                    else:
                        log.warning(
                            "%r defines alias %r pointing to existing page %r",
                            page,
                            alias,
                            old,
                        )
                    continue
                node.create_auto_page_as_path(
                    page_cls=AliasPage,
                    created_from=page,
                    page=page,
                    title=page.title,
                    name=path.name,
                )


class AliasPage(TemplatePage, AutoPage):
    """
    Page rendering a redirect to another page
    """

    page = SourcePageField(doc="Page this alias redirects to")
    TYPE = "alias"
    TEMPLATE = "redirect.html"

    def _compute_change_extent(self) -> ChangeExtent:
        if self.created_from is None:
            raise AssertionError(f"AliasPage {self!r} has empty created_from")
        res = self.created_from.change_extent
        if res == ChangeExtent.CONTENTS:
            res = ChangeExtent.UNCHANGED
        return res


FEATURES = {
    "aliases": AliasesFeature,
}



from __future__ import annotations

from collections.abc import Iterable, Mapping
from graphlib import CycleError, TopologicalSorter
from typing import TypeVar

__all__ = ["CycleError", "sort"]

N = TypeVar("N")


def sort(graph: Mapping[N, Iterable[N]]) -> list[N]:
    """
    Linearize a dependency graph, throwing an exception if a cycle is detected.

    When no dependencies intervene in ordering, the algorithm preserves the
    original insertion order of the graph.

    :arg graph: a dict mapping each node to an iterable of their antecedents
    """
    sorter = TopologicalSorter(graph)
    return list(sorter.static_order())



from unittest import TestCase

from . import utils as test_utils


class MockContext(dict):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.name = "test"


class TestUrlFor(test_utils.MockSiteTestMixin, TestCase):
    """
    Test theme functions
    """

    def test_no_site_root(self):
        files = {
            ".staticsite": {
                "site_url": "https://www.example.org",
            },
            "page1.md": {},
            "dir/page2.md": {},
            "dir/index.html": {},
        }

        with self.site(files) as mocksite:
            site = mocksite.site

            def url_for(dest, page=None, **kw):
                if page:
                    context = MockContext(page=page)
                else:
                    context = None
                return site.theme.jinja2_url_for(context, dest, **kw)

            page = mocksite.page("")
            self.assertEqual(url_for("page1.md", page=page), "/page1")
            self.assertEqual(url_for("page1", page=page), "/page1")

            # Autogenerated index.html can be looked up now
            self.assertEqual(url_for("page1/index.html", page=page), "/page1")

            # index.html however resolve, as they exist in the sources
            # namespace
            self.assertEqual(url_for("dir", page=page), "/dir")
            self.assertEqual(url_for("dir/index.html", page=page), "/dir")

            # Test absolute urls
            self.assertEqual(
                url_for("page1", page=page, absolute=True),
                "https://www.example.org/page1",
            )

    def test_site_path(self):
        files = {
            ".staticsite": {
                "site_url": "https://www.example.org",
                "site_path": "prefix",
            },
            "page1.md": {},
            "page2.rst": {},
            "page3.yaml": {"data_type": "page"},
            "dir/page2.md": {},
            "dir/index.html": {},
        }

        with self.site(files) as mocksite:
            site = mocksite.site

            def url_for(dest, page=None, **kw):
                if page:
                    context = MockContext(page=page)
                else:
                    context = None
                return site.theme.jinja2_url_for(context, dest, **kw)

            page = mocksite.page("")
            self.assertEqual(url_for("page1.md", page=page), "/prefix/page1")
            self.assertEqual(url_for("page1", page=page), "/prefix/page1")

            self.assertEqual(url_for("page2.rst", page=page), "/prefix/page2")
            self.assertEqual(url_for("page2", page=page), "/prefix/page2")

            self.assertEqual(url_for("page3.yaml", page=page), "/prefix/page3")
            self.assertEqual(url_for("page3", page=page), "/prefix/page3")

            # Test absolute urls
            self.assertEqual(
                url_for("page1", page=page, absolute=True),
                "https://www.example.org/prefix/page1",
            )


class TestMarkdownFilter(test_utils.MockSiteTestMixin, TestCase):
    def test_markdown(self):
        files = {
            "index.html": "",
        }
        with self.site(files) as mocksite:
            site = mocksite.site

            page = mocksite.page("")

            tpl = site.theme.jinja2.from_string(
                "{% filter markdown %}*This* is an [example](http://example.org){% endfilter %}"
            )

            res = tpl.render(page=page)
            self.assertEqual(
                res,
                '<p><em>This</em> is an <a href="http://example.org">example</a></p>',
            )



from __future__ import annotations

import datetime
import logging
import os
import re
from collections import defaultdict
from collections.abc import Callable, MutableMapping, Sequence
from typing import TYPE_CHECKING, Any

import jinja2
import jinja2.sandbox
import markupsafe

from . import toposort
from .file import File
from .page import ImagePage, Page, PageNotFoundError
from .utils import front_matter
from .utils.arrange import arrange

if TYPE_CHECKING:
    from .page import Pages
    from .site import Site

log = logging.getLogger("theme")


class ThemeNotFoundError(Exception):
    pass


class Loader:
    """
    Theme loader, resolving theme dependency chains
    """

    def __init__(self, search_paths: Sequence[str]):
        # Sequence of search paths to use to resolve theme names
        self.search_paths = search_paths
        # Configurations by name
        self.configs: dict[str, dict[str, Any]] = {}
        # Dependency graph of themes
        self.deps: dict[str, set[str]] = defaultdict(set)

    def load(self, name: str) -> list[dict[str, Any]]:
        """
        Load the configuration of the given theme and all its dependencies.

        Return the list of all the resulting configurations in topological
        order, with the bottom-most dependency first and the named theme last
        """
        self.load_configs(name)
        sorted_names = toposort.sort(self.deps)
        return [self.configs[name] for name in sorted_names]

    def load_legacy(self, path: str) -> list[dict[str, Any]]:
        """
        Same as load, but start with a path to an initial theme
        """
        name = os.path.basename(path)
        config = self.load_config(path, name)
        self.configs[name] = config
        self.deps[name]

        for parent in config["extends"]:
            self.load_configs(parent)
            self.deps[name].add(parent)

        sorted_names = toposort.sort(self.deps)
        return [self.configs[name] for name in sorted_names]

    def load_configs(self, name: str) -> None:
        """
        Populate self.configs with the configuration of the named theme,
        preceded by all its dependencies
        """
        if name in self.configs:
            log.warning(
                "%s: dependency loop found between themes: ignoring dependency", name
            )
            return

        config = self.load_config(self.find_root(name), name)
        self.configs[name] = config
        self.deps[name]

        for parent in config["extends"]:
            self.load_configs(parent)
            self.deps[name].add(parent)

    def find_root(self, name: str) -> str:
        """
        Lookup the root directory of a theme by name
        """
        for path in self.search_paths:
            root = os.path.join(path, name)
            if os.path.isdir(root):
                return root

        raise ThemeNotFoundError(f"Theme {name!r} not found in {self.search_paths!r}")

    def load_config(self, root: str, name: str) -> dict[str, Any]:
        """
        Load the configuration for the given named theme
        """
        pathname = os.path.join(root, "config")
        if not os.path.isfile(pathname):
            config: dict[str, Any] = {}
        else:
            with open(pathname) as fd:
                fmt, config = front_matter.read_whole(fd)

        # Normalize 'extends' to a list of strings
        extends = config.get("extends")
        if extends is None:
            config["extends"] = []
        elif isinstance(extends, str):
            config["extends"] = [extends]

        # Theme name
        config["name"] = name

        # Absolute path to the root of the theme directory
        config["root"] = os.path.abspath(root)

        return config


class Jinja2TemplateLoader(jinja2.loaders.BaseLoader):
    re_content = re.compile(r"^content:(.+)")

    def __init__(self, theme: Theme):
        self.loader_content = jinja2.FileSystemLoader(theme.site.content_root)
        self.loader_theme = jinja2.FileSystemLoader(theme.template_lookup_paths)

    def get_loader(self, template: str) -> tuple[jinja2.loaders.BaseLoader, str]:
        mo = self.re_content.match(template)
        if mo:
            return self.loader_content, mo.group(1)
        else:
            return self.loader_theme, template

    def get_source(
        self, environment: jinja2.Environment, template: str
    ) -> tuple[str, str | None, Callable[[], bool] | None]:
        loader, name = self.get_loader(template)
        try:
            return loader.get_source(environment, name)
        except jinja2.TemplateNotFound:
            # re-raise the exception with the correct filename here.
            # (the one that includes the prefix)
            raise jinja2.TemplateNotFound(template)

    @jinja2.utils.internalcode
    def load(
        self,
        environment: jinja2.Environment,
        name: str,
        globals: MutableMapping[str, Any] | None = None,
    ) -> jinja2.Template:
        loader, local_name = self.get_loader(name)
        try:
            return loader.load(environment, local_name, globals)
        except jinja2.TemplateNotFound:
            # re-raise the exception with the correct filename here.
            # (the one that includes the prefix)
            raise jinja2.TemplateNotFound(name)

    def list_templates(self) -> list[str]:
        result = []
        for template in self.loader_theme.list_templates():
            result.append(template)
        for template in self.loader_content.list_templates():
            result.append("content:" + template)
        return result


class Theme:
    def __init__(self, site: Site, name: str, configs: list[dict[str, Any]]):
        # Site object
        self.site = site

        # Template name
        self.name = name

        # Configuration for this theme and all the dependencies, sorted from
        # the earliest dependency to the theme
        self.configs = configs

        # Jinja2 Environment
        self.jinja2: jinja2.Environment

        # Compute template lookup paths
        self.template_lookup_paths: list[str] = []
        for config in reversed(self.configs):
            self.template_lookup_paths.append(config["root"])
        log.info("%s: template lookup paths: %r", self.name, self.template_lookup_paths)

        # Compute feature directories
        self.feature_dirs = []
        for config in self.configs:
            features_dir = os.path.join(config["root"], "features")
            if os.path.isdir(features_dir):
                self.feature_dirs.append(features_dir)
        log.info("%s: feature directories: %r", self.name, self.feature_dirs)

        # Compute system asset names
        self.system_assets = set(self.site.settings.SYSTEM_ASSETS)
        for config in self.configs:
            self.system_assets.update(config.get("system_assets", ()))
        log.info("%s: system assets: %r", self.name, self.system_assets)

        # Compute theme static asset dirs
        self.theme_static_dirs = []
        for config in self.configs:
            theme_static = os.path.join(config["root"], "static")
            if os.path.isdir(theme_static):
                self.theme_static_dirs.append(theme_static)
        log.info("%s: theme static directories: %r", self.name, self.theme_static_dirs)

        # Merge theme metadata
        meta_keys = frozenset(("image_sizes",))
        self.meta: dict[str, Any] = {}
        for config in self.configs:
            for key in config.keys() & meta_keys:
                self.meta[key] = config[key]

    @classmethod
    def create(
        cls, site: Site, name: str, search_paths: Sequence[str] | None = None
    ) -> Theme:
        """
        Create a Theme looking up its name in the theme search paths
        """
        if site.settings.PROJECT_ROOT is None:
            raise RuntimeError("PROJECT_ROOT is None")

        if search_paths is None:
            search_paths = [
                os.path.join(site.settings.PROJECT_ROOT, path)
                for path in site.settings.THEME_PATHS
            ]
        loader = Loader(search_paths)
        return cls(site, name, loader.load(name))

    @classmethod
    def create_legacy(cls, site: Site, paths: Sequence[str]) -> Theme:
        """
        Create a theme from a list of possible theme paths
        """
        if site.settings.PROJECT_ROOT is None:
            raise RuntimeError("PROJECT_ROOT is None")

        loader = Loader(site.settings.THEME_PATHS)

        for root in paths:
            root = os.path.join(site.settings.PROJECT_ROOT, root)
            if os.path.isdir(root):
                return cls(site, os.path.basename(root), loader.load_legacy(root))

        raise ThemeNotFoundError(f"Theme not found in {paths!r}")

    def load(self) -> None:
        # Load feature plugins from the theme directories
        self.site.features.load_feature_dir(self.feature_dirs)
        self.site.features.commit()
        self.site.stage_features_constructed = True

        env_cls: type[jinja2.Environment]
        # Jinja2 template engine
        if self.site.settings.JINJA2_SANDBOXED:
            env_cls = jinja2.sandbox.ImmutableSandboxedEnvironment
        else:
            env_cls = jinja2.Environment

        self.jinja2 = env_cls(
            loader=Jinja2TemplateLoader(self),
            autoescape=True,
        )

        # Add settings to jinja2 globals
        for x in dir(self.site.settings):
            if not x.isupper():
                continue
            self.jinja2.globals[x] = getattr(self.site.settings, x)

        self.jinja2.globals["site"] = self.site
        self.jinja2.globals["regex"] = re.compile

        # Install site's functions into the jinja2 environment
        self.jinja2.globals.update(
            has_page=self.jinja2_has_page,
            url_for=self.jinja2_url_for,
            page_for=self.jinja2_page_for,
            site_pages=self.jinja2_site_pages,
            img_for=self.jinja2_img_for,
            now=self.site.generation_time,
            next_month=(
                self.site.generation_time.replace(day=1) + datetime.timedelta(days=40)
            ).replace(day=1, hour=0, minute=0, second=0, microsecond=0),
        )

        self.jinja2.filters["datetime_format"] = self.jinja2_datetime_format
        self.jinja2.filters["next_month"] = self.jinja2_next_month
        self.jinja2.filters["basename"] = self.jinja2_basename
        self.jinja2.filters["arrange"] = self.jinja2_arrange

        # Add feature-provided globals and filters
        for feature in self.site.features.ordered():
            self.jinja2.globals.update(feature.j2_globals)
            self.jinja2.filters.update(feature.j2_filters)

    def scan_assets(self) -> None:
        """
        Load static assets
        """
        # Load system assets from site settings and theme configurations
        for name in self.system_assets:
            root = os.path.join("/usr/share/javascript", name)
            if not os.path.isdir(root):
                log.warning("%s: system asset directory not found", root)
                continue
            src = File.with_stat(name, root)
            self.site.scan_asset_tree(
                src=src,
                node=self.site.static_root.asset_child(name, src=src),
            )

        # Load assets from theme directories
        for path in self.theme_static_dirs:
            self.site.scan_asset_tree(
                src=File.with_stat("", os.path.abspath(path)),
                node=self.site.static_root,
            )

    def jinja2_basename(self, val: str) -> str:
        return os.path.basename(val)

    @jinja2.pass_context
    def jinja2_datetime_format(
        self,
        context: jinja2.runtime.Context,
        dt: str | datetime.datetime,
        format: str | None = None,
    ) -> str:
        if not isinstance(dt, datetime.datetime):
            import dateutil.parser

            dt = dateutil.parser.parse(dt)
        if format is None or format == "iso8601":
            from .utils import format_date_iso8601

            return format_date_iso8601(dt)
        elif format in ("rss2", "rfc822"):
            from .utils import format_date_rfc822

            return format_date_rfc822(dt)
        elif format in ("atom", "rfc3339"):
            from .utils import format_date_rfc3339

            return format_date_rfc3339(dt)
        elif format == "w3cdtf":
            from .utils import format_date_w3cdtf

            return format_date_w3cdtf(dt)
        elif format[0] == "%":
            return dt.strftime(format)
        else:
            log.warning(
                "%s+%s: invalid datetime format %r requested",
                context.parent["page"].src.relpath,
                context.name,
                format,
            )
            return f"(unknown datetime format {format})"

    @jinja2.pass_context
    def jinja2_next_month(
        self, context: jinja2.runtime.Context, dt: Any
    ) -> datetime.date | datetime.datetime:
        if isinstance(dt, str):
            import dateutil.parser

            dt = dateutil.parser.parse(dt)

        if isinstance(dt, datetime.datetime):
            return (dt.replace(day=1) + datetime.timedelta(days=40)).replace(
                day=1, hour=0, minute=0, second=0, microsecond=0
            )
        elif isinstance(dt, datetime.date):
            return (dt.replace(day=1) + datetime.timedelta(days=40)).replace(day=1)
        else:
            raise RuntimeError(
                f"{context.parent['page']}+{context.name}: invalid datetime {dt!r} of type {type(dt)}:"
                " accepted are str, datetime, date"
            )

    @jinja2.pass_context
    def jinja2_has_page(self, context: jinja2.runtime.Context, arg: str) -> bool:
        cur_page = context.get("page")
        try:
            cur_page.resolve_path(arg)
        except PageNotFoundError:
            return False
        else:
            return True

    @jinja2.pass_context
    def jinja2_page_for(self, context: jinja2.runtime.Context, arg: str | Page) -> Page:
        """
        Generate a URL for a page, specified by path or with the page itself
        """
        if isinstance(arg, Page):
            return arg

        cur_page: Page | None = context.get("page")
        if cur_page is None:
            raise RuntimeError(
                f"{cur_page}+{context.name}: page_for({arg!r}): current page is not defined"
            )

        try:
            return cur_page.resolve_path(arg)
        except PageNotFoundError as e:
            raise RuntimeError(f"{cur_page}+{context.name}: {e}")

    @jinja2.pass_context
    def jinja2_url_for(
        self,
        context: jinja2.runtime.Context,
        arg: str | Page,
        absolute: bool = False,
        static: bool = False,
    ) -> str:
        """
        Generate a URL for a page, specified by path or with the page itself
        """
        cur_page: Page | None = context.get("page")
        # print(f"Theme.jinja2_url_for {cur_page=!r}")
        if cur_page is None:
            log.warning(
                "%s+%s: url_for(%s): current page is not defined",
                cur_page,
                context.name,
                arg,
            )
            return ""

        try:
            return cur_page.url_for(arg, absolute=absolute, static=static)
        except PageNotFoundError as e:
            log.warning("%s:%s: %s", cur_page, context.name, e)
            return ""

    @jinja2.pass_context
    def jinja2_site_pages(
        self, context: jinja2.runtime.Context, **kw: Any
    ) -> list[Page]:
        cur_page: Page | None = context.get("page")
        if cur_page is None:
            log.warning(
                "%s+%s: site_pages: current page is not defined", cur_page, context.name
            )
            return []

        return cur_page.find_pages(**kw)

    @jinja2.pass_context
    def jinja2_img_for(
        self,
        context: jinja2.runtime.Context,
        path: str | Page,
        type: str | None = None,
        absolute: bool = False,
        **attrs: Any,
    ) -> str:
        cur_page = context.get("page")
        if cur_page is None:
            log.warning(
                "%s+%s: img(%s): current page is not defined",
                cur_page,
                context.name,
                path,
            )
            return ""

        image_page = cur_page.resolve_path(path)
        if isinstance(image_page, ImagePage):
            res_attrs = image_page.get_img_attributes(type=type, absolute=absolute)
        else:
            log.warning(
                "%s: img src= resolves to %s which is not an image page",
                cur_page,
                image_page,
            )
            res_attrs = {"src": cur_page.url_for(image_page)}

        escape = markupsafe.escape
        res = ["<img"]
        for k, v in res_attrs.items():
            res.append(f" {escape(k)}='{escape(v)}'")
        res.append("></img>")
        return markupsafe.Markup("".join(res))

    def jinja2_arrange(
        self, pages: Pages | list[Page], *args: Any, **kw: Any
    ) -> list[Page]:
        from .page import Pages

        if isinstance(pages, Pages):
            return pages.arrange(*args, **kw)
        else:
            return arrange(pages, *args, **kw)



from __future__ import annotations

from typing import Any, Optional, Union

IN_CLOSE_WRITE: int
IN_DELETE: int


class WatchManager:
    def __init__(self) -> None:
        ...

    def add_watch(
            self, path: Union[str, list[str]], mask: int, proc_fun: Any = None, rec: bool = False,
            auto_add: bool = False, do_glob: bool = False, quiet: bool = True,
            exclude_filter: Any = None) -> dict[str, int]:
        ...

    def rm_watch(self, wd: Union[int, list[int]], rec: bool = False, quiet: bool = True) -> dict[int, bool]:
        ...


class Watch:
    ...


class Event:
    name: str
    path: str


class AsyncioNotifier:
    def __init__(
            self, watch_manager: WatchManager, loop: Any, callback: Any = None,
            default_proc_fun: Any = None, read_freq: int = 0, threshold: int = 0, timeout: Optional[int] = None):
        ...



from __future__ import annotations

import datetime
import logging
from typing import TYPE_CHECKING, Any, TypeVar

from . import fields, file
from .node import Node, SkipPage
from .site import Path

if TYPE_CHECKING:
    from .asset import Asset
    from .page import Page
    from .site import Site

log = logging.getLogger("node")

P = TypeVar("P", bound="Page")


class SourceNode(Node):
    """
    Node corresponding to a source directory
    """

    def add_asset(self, *, src: file.File, name: str) -> Asset | None:
        """
        Add an Asset as a subnode of this one
        """
        # Import here to avoid cyclical imports
        from .asset import Asset

        return self.create_source_page_as_file(
            page_cls=Asset, src=src, name=name, dst=name
        )

    def create_source_page_as_file(
        self,
        *,
        page_cls: type[P],
        src: file.File,
        dst: str,
        date: datetime.datetime | None = None,
        **kw: Any,
    ) -> P | None:
        """
        Create a page of the given type, attaching it at the given path
        """
        if "created_from" in kw:
            raise RuntimeError("source page created with 'created_from' set")

        if self.site.last_load_step > self.site.LOAD_STEP_CONTENTS:
            raise RuntimeError(
                "Node.create_source_page created after the 'contents' step has completed"
            )

        if self.site.previous_source_footprints:
            # TODO: since we pop, we lose fooprint info for assets that replace other assets.
            # TODO: propagate it when doing the replacement?
            kw["old_footprint"] = self.site.previous_source_footprints.pop(
                src.relpath, None
            )

        if date is None:
            date = self.site.localized_timestamp(src.stat.st_mtime)
        else:
            date = self.site.clean_date(date)

        # Skip draft pages
        if date > self.site.generation_time and not self.site.settings.DRAFT_MODE:
            log.info("%s: page is still a draft: skipping", src.relpath)
            return None

        try:
            return self._create_leaf_page(
                page_cls=page_cls, dst=dst, src=src, date=date, **kw
            )
        except SkipPage:
            return None

    def asset_child(self, name: str, src: file.File) -> SourceAssetNode:
        """
        Create the given child as a SourceNode
        """
        if (node := self.sub.get(name)) is not None:
            if not isinstance(node, SourceAssetNode):
                raise RuntimeError(
                    f"source node {name} already exists in {self.name}"
                    f" as {node.__class__.__name__} instead of SourceAssetNode {src.abspath}"
                )
            if src not in node.srcs:
                log.debug("assets: merging %s into %s", src.abspath, node)
                node.srcs.append(src)
            return node
        else:
            node = self.site.features.get_node_class(SourceAssetNode)(
                site=self.site, name=name, parent=self, src=src
            )
            self.sub[name] = node
            return node


class SourceAssetNode(SourceNode):
    """
    Node corresponding to a source directory that contains only asset pages
    """

    def __init__(self, site: Site, name: str, *, src: file.File, parent: Node):
        super().__init__(site, name, parent=parent)
        self.srcs: list[file.File] = [src]


class SourcePageNode(SourceNode):
    """
    Node corresponding to a source directory that contains non-asset pages
    """

    def __init__(self, site: Site, name: str, *, src: file.File, parent: Node | None):
        super().__init__(site, name, parent=parent)
        self.src: file.File = src

    def create_source_page_as_path(
        self,
        page_cls: type[P],
        src: file.File,
        name: str,
        date: datetime.datetime | None = None,
        **kw: Any,
    ) -> P | None:
        """
        Create a page of the given type, attaching it at the given path
        """
        if "created_from" in kw:
            raise RuntimeError("source page created with 'created_from' set")

        if self.site.last_load_step > self.site.LOAD_STEP_CONTENTS:
            raise RuntimeError(
                "Node.create_source_page created after the 'contents' step has completed"
            )

        if self.site.previous_source_footprints:
            # TODO: since we pop, we lose fooprint info for assets that replace other assets.
            # TODO: propagate it when doing the replacement?
            kw["old_footprint"] = self.site.previous_source_footprints.pop(
                src.relpath, None
            )

        if date is None:
            date = self.site.localized_timestamp(src.stat.st_mtime)
        else:
            date = self.site.clean_date(date)

        # Skip draft pages
        if date > self.site.generation_time and not self.site.settings.DRAFT_MODE:
            log.info("%s: page is still a draft: skipping", src.relpath)
            return None

        node = self._child(name)

        try:
            return node._create_index_page(
                page_cls=page_cls, directory_index=False, src=src, date=date, **kw
            )
        except SkipPage:
            return None

    def create_source_page_as_index(
        self,
        page_cls: type[P],
        src: file.File,
        date: datetime.datetime | None = None,
        **kw: Any,
    ) -> P | None:
        """
        Create a page of the given type, attaching it at the given path
        """
        if "created_from" in kw:
            raise RuntimeError("source page created with 'created_from' set")

        if self.site.last_load_step > self.site.LOAD_STEP_CONTENTS:
            raise RuntimeError(
                "Node.create_source_page created after the 'contents' step has completed"
            )

        if self.site.previous_source_footprints:
            # TODO: since we pop, we lose fooprint info for assets that replace other assets.
            # TODO: propagate it when doing the replacement?
            kw["old_footprint"] = self.site.previous_source_footprints.pop(
                src.relpath, None
            )

        if date is None:
            date = self.site.localized_timestamp(src.stat.st_mtime)
        else:
            date = self.site.clean_date(date)

        # Skip draft pages
        if date > self.site.generation_time and not self.site.settings.DRAFT_MODE:
            log.info("%s: page is still a draft: skipping", src.relpath)
            return None

        try:
            return self._create_index_page(
                page_cls=page_cls, directory_index=True, src=src, date=date, **kw
            )
        except SkipPage:
            return None

    def page_child(self, name: str, src: file.File) -> SourcePageNode:
        """
        Create the given child as a SourceNode
        """
        if (node := self.sub.get(name)) is not None:
            if not isinstance(node, SourcePageNode):
                raise RuntimeError(
                    f"source node {name} already exists in {self.name}"
                    f" as {node.__class__.__name__} instead of source node {src.abspath}"
                )
            if node.src != src:
                raise RuntimeError(
                    f"source node {name} already exists in {self.name}"
                    f" as {node.src.abspath} instead of {src.abspath}"
                )
            return node

        node = self.site.features.get_node_class(SourcePageNode)(
            site=self.site, name=name, parent=self, src=src
        )
        self.sub[name] = node
        return node


class RootNode(SourcePageNode):
    """
    Node at the root of the site tree
    """

    title = fields.Str["Node"](
        doc="""
        Title used as site name.

        This only makes sense for the root node of the site hierarchy, and
        takes the value from the title of the root index page. If set, and the
        site name is not set by other means, it is used to give the site a name.
    """
    )

    def __init__(self, site: Site, *, src: file.File):
        super().__init__(site, name="", src=src, parent=None)

    def static_root(self, path: Path) -> SourceNode:
        """
        Return the subnode at the given path, creating it if missing
        """
        res: SourceNode = self
        while path:
            res = res.asset_child(path.head, src=self.src)
            path = path.tail
        return res

    def generate_path(self, path: Path) -> Node:
        """
        Return the subnode at the given path, creating it if missing
        """
        res: Node = self
        while path:
            res = res._child(path.head)
            path = path.tail
        return res



import re
from unittest import TestCase

from staticsite.page_filter import PageFilter

from . import utils as test_utils


def select(mocksite, *args, **kw):
    f = PageFilter(mocksite.site, *args, **kw)
    return [p.site_path for p in f.filter()]


class TestPageFilter(test_utils.MockSiteTestMixin, TestCase):
    def test_site(self):
        files = {
            # taxonomies are not findable pages
            "taxonomies/tags.taxonomy": {},
            "page.md": {},
            "blog/post1.md": {},
            "blog/post2.md": {},
        }
        with self.site(files) as mocksite:
            mocksite.assertPagePaths(
                (
                    "",
                    "page",
                    "taxonomies/tags",
                    "blog",
                    "blog/post1",
                    "blog/post2",
                    "taxonomies",
                )
            )

            self.assertCountEqual(
                select(mocksite, path="blog/*"),
                [
                    "blog/post1",
                    "blog/post2",
                ],
            )

            self.assertCountEqual(
                select(mocksite, path=r"^blog/"),
                [
                    "blog/post1",
                    "blog/post2",
                ],
            )

            self.assertEqual(
                select(mocksite, path=re.compile(r"^[tp]")),
                [
                    "page",
                ],
            )

    def test_relative(self):
        files = {
            "page.md": {"pages": "dir/*"},
            "dir/page1.md": {"pages": "dir/*"},
            "dir/page2.md": {},
            "dir/dir/page3.md": {},
            "dir/dir/page4.md": {},
        }
        with self.site(files) as mocksite:
            mocksite.assertPagePaths(
                (
                    "",
                    "page",
                    "dir",
                    "dir/page1",
                    "dir/page2",
                    "dir/dir",
                    "dir/dir/page3",
                    "dir/dir/page4",
                )
            )

            page, page1, page2, page3, page4 = mocksite.page(
                "page", "dir/page1", "dir/page2", "dir/dir/page3", "dir/dir/page4"
            )
            self.assertCountEqual(page.pages, [page1, page2, page3, page4])
            self.assertCountEqual(page1.pages, [page3, page4])



from __future__ import annotations

import datetime
import os
import time
from contextlib import contextmanager
from unittest import TestCase

import pytz

from staticsite.file import File
from staticsite.site import Site

from . import utils as test_utils


@contextmanager
def override_tz(val):
    orig = os.environ.get("TZ", None)
    os.environ["TZ"] = val
    time.tzset()
    yield
    if orig is None:
        del os.environ["TZ"]
    else:
        os.environ["TZ"] = orig
    time.tzset()


class TestAsset(test_utils.MockSiteTestMixin, TestCase):
    @test_utils.assert_no_logs()
    @override_tz("Pacific/Samoa")
    def test_timestamps(self):
        # $ TZ=UTC date +%s --date="2016-11-01" → 1477958400
        FILE_TS = 1477958400

        sitedef = test_utils.MockSite(
            {
                ".staticsite": {
                    "asset": True,
                },
                "testasset": "test",
            }
        )
        sitedef.auto_load_site = False
        sitedef.mock_file_mtime = None

        with self.site(sitedef) as mocksite:
            # Set file mtime and check that it can be read back correctly
            path = os.path.join(mocksite.root, "testasset")
            os.utime(path, (FILE_TS, FILE_TS))
            src = File.with_stat("testasset", path)
            self.assertEqual(src.stat.st_mtime, FILE_TS)

            # Load site contents
            mocksite.load_site(until=Site.LOAD_STEP_CONTENTS)

            page = mocksite.page("testasset")

            self.assertEqual(page.node.name, "")
            self.assertEqual(page.TYPE, "asset")
            self.assertEqual(page.src.stat.st_mtime, FILE_TS)
            self.assertEqual(
                page.meta["date"],
                datetime.datetime(2016, 11, 1, 0, 0, 0, tzinfo=pytz.utc),
            )
            self.assertEqual(page.meta["site_url"], "https://www.example.org")
            self.assertEqual(page.site_path, "testasset")
            self.assertEqual(page.node.path, "")
            self.assertEqual(page.dst, "testasset")
            self.assertFalse(page.directory_index)



from unittest import TestCase

from . import utils as test_utils


class TestDirs(test_utils.MockSiteTestMixin, TestCase):
    """
    Test dirs feature
    """

    def test_dirs(self):
        files = {
            "page_root.md": {},
            "dir1/page_sub.md": {},
            "dir1/dir2/dir3/page_sub3.md": {},
            "dir1/dir2/dir3/page_sub3/slides.pdf": "",
            "dir1/dir2/dir3/page_sub4.md": {},
        }
        with self.site(files) as mocksite:
            # We have a root dir index and dir indices for all subdirs
            dir_root, dir_dir1, dir_dir2, dir_dir3 = mocksite.page(
                "", "dir1", "dir1/dir2", "dir1/dir2/dir3"
            )

            self.assertEqual(dir_root.TYPE, "dir")
            self.assertEqual(dir_dir1.TYPE, "dir")
            self.assertEqual(dir_dir3.TYPE, "dir")
            self.assertEqual(dir_dir3.TYPE, "dir")

            # Check the contents of all dirs
            self.assertCountEqual(dir_root.meta["pages"], [mocksite.page("page_root")])
            self.assertCountEqual(dir_root.subdirs, [dir_dir1])
            self.assertIsNone(dir_root.parent)
            self.assertEqual(dir_root.meta["title"], "Test site")
            self.assertEqual(dir_root.meta["template"], "dir.html")

            self.assertCountEqual(
                dir_dir1.meta["pages"], [mocksite.page("dir1/page_sub")]
            )
            self.assertCountEqual(dir_dir1.subdirs, [dir_dir2])
            self.assertEqual(dir_dir1.parent, dir_root)
            self.assertEqual(dir_dir1.meta["title"], "dir1")

            self.assertCountEqual(dir_dir2.meta["pages"], [])
            self.assertCountEqual(dir_dir2.subdirs, [dir_dir3])
            self.assertEqual(dir_dir2.parent, dir_dir1)
            self.assertEqual(dir_dir2.meta["title"], "dir2")

            self.assertCountEqual(
                dir_dir3.meta["pages"],
                [
                    mocksite.page("dir1/dir2/dir3/page_sub3"),
                    mocksite.page("dir1/dir2/dir3/page_sub4"),
                ],
            )
            self.assertCountEqual(dir_dir3.subdirs, [])
            self.assertEqual(dir_dir3.parent, dir_dir2)
            self.assertEqual(dir_dir3.meta["title"], "dir3")

            asset = mocksite.page("dir1/dir2/dir3/page_sub3/slides.pdf")
            self.assertEqual(asset.TYPE, "asset")



from __future__ import annotations

from unittest import TestCase

from . import utils as test_utils


class TestLoad(test_utils.MockSiteTestMixin, TestCase):
    def test_leaf_nodes(self):
        """
        Test that empty leaf nodes are pruned
        """
        files = {
            "index.md": {},
            "drafts/index.md": {"date": "2040-01-01"},
            "skipped/.staticsite": {"skip": True},
            "dir1/dir2/skipped/.staticsite": {"skip": True},
            "dir3/dir4/drafts/index.md": {"date": "2040-01-01"},
            "empty/.gitignore": "",
            "dir5/dir6/empty/.gitignore": "",
        }
        with self.site(files) as mocksite:
            mocksite.assertPagePaths(("",))

    def test_ignore(self):
        files = {
            "index.md": {},
            "index.md.swp": {},
            "index.md~": {},
            ".staticsite": {
                "dirs": {"assets": {"asset": True}},
                "ignore": ["*.swp", "*~"],
            },
            "drafts/index.md": {},
            "drafts/index.md~": {},
            "drafts/index.md.swp": {},
            "assets/file.txt": "",
            "assets/file.txt.swp": "",
            "assets/file.txt~": "",
            "assets/sub/file1.txt": "",
            "assets/sub/file1.txt.swp": "",
            "assets/sub/file1.txt~": "",
        }
        with self.site(files) as mocksite:
            mocksite.assertPagePaths(
                (
                    "",
                    "drafts",
                    "assets/file.txt",
                    "assets/sub/file1.txt",
                )
            )



from __future__ import annotations

import argparse
import logging
import os
import shlex
import subprocess
from typing import TYPE_CHECKING, Any

from .command import Fail, SiteCommand, register

if TYPE_CHECKING:
    from ..site import Site

log = logging.getLogger("new")


class LazyTitle:
    """
    Wrapper for the value of a post title.

    If no value has been provided and one is requested by the template, ask the
    user for one.
    """

    def __init__(self, title: str | None = None) -> None:
        self.title = title

    def __str__(self) -> str:
        if self.title is None:
            self.title = input("Please enter the post title: ")
        return self.title

    def __len__(self) -> int:
        return len(str(self))


class LazySlug:
    """
    Generate a slug from the value of LazyTitle
    """

    def __init__(self, site: Site, lazy_title: LazyTitle):
        self.site = site
        self.lazy_title = lazy_title

    def __str__(self) -> str:
        return self.site.slugify(str(self.lazy_title))


@register
class New(SiteCommand):
    "create a new page"

    def run(self) -> int | None:
        site = self.load_site()

        archetype = site.archetypes.find(self.args.archetype)
        if archetype is None:
            raise Fail(f"archetype {self.args.archetype} not found")
        log.info("Using archetype %s", archetype.relpath)

        title = LazyTitle(self.args.title)
        slug = LazySlug(site, title)
        meta, body = archetype.render(slug=slug, title=title)

        relpath = meta.get("path", None)
        if relpath is None:
            raise Fail(
                f"archetype {archetype.relpath} does not contain `path` in its front matter"
            )

        if site.settings.PROJECT_ROOT is None:
            raise Fail("PROJECT_ROOT is empty")
        if site.settings.CONTENT is None:
            raise Fail("CONTENT is empty")

        content_root = os.path.join(site.settings.PROJECT_ROOT, site.settings.CONTENT)
        abspath = os.path.join(content_root, relpath)
        if self.args.overwrite or not os.path.exists(abspath):
            os.makedirs(os.path.dirname(abspath), exist_ok=True)
            with open(abspath, "wt") as out:
                out.write(body)
        else:
            log.info("%s already exists: reusing it", abspath)

        if not self.args.noedit:
            settings_dict = site.settings.as_dict()
            cmd = [
                x.format(name=abspath, slug=slug, **settings_dict)
                for x in site.settings.EDIT_COMMAND
            ]
            try:
                subprocess.check_call(cmd)
            except subprocess.CalledProcessError as e:
                log.warning(
                    "Editor command %s exited with error %d",
                    " ".join(shlex.quote(x) for x in cmd),
                    e.returncode,
                )
                return e.returncode
        print(abspath)
        return None

    @classmethod
    def add_subparser(
        cls, subparsers: argparse._SubParsersAction[Any]
    ) -> argparse.ArgumentParser:
        parser = super().add_subparser(subparsers)
        parser.add_argument(
            "-a", "--archetype", default="default", help="page archetype"
        )
        parser.add_argument("-t", "--title", help="page title")
        parser.add_argument(
            "-n",
            "--noedit",
            action="store_true",
            help="do not run an editor, only output the file name of the new post",
        )
        parser.add_argument(
            "--overwrite",
            action="store_true",
            help="if a post already exists, overwrite it instead of reusing it",
        )
        return parser



from __future__ import annotations

import os
from unittest import TestCase

from . import utils as test_utils


class TestRst(test_utils.MockSiteTestMixin, TestCase):
    def test_simple(self):
        self.maxDiff = None

        files = {
            "taxonomies/tags.taxonomy": {},
            "index.rst": """
:date: 2016-04-16 10:23:00+02:00
:tags: [example, "another_tag"]
:description:
  line1
  line2
  line3

Example blog post in reStructuredText
=====================================
""",
        }

        with self.site(files) as mocksite:
            mocksite.assertPagePaths(
                (
                    "",
                    "taxonomies",
                    "taxonomies/tags",
                    "taxonomies/tags/example",
                    "taxonomies/tags/example/index.rss",
                    "taxonomies/tags/example/index.atom",
                    "taxonomies/tags/example/archive",
                    "taxonomies/tags/another_tag",
                    "taxonomies/tags/another_tag/index.rss",
                    "taxonomies/tags/another_tag/index.atom",
                    "taxonomies/tags/another_tag/archive",
                )
            )
            page, tags, tag_example, tag_another = mocksite.page(
                "",
                "taxonomies/tags",
                "taxonomies/tags/example",
                "taxonomies/tags/another_tag",
            )

            # We have a root dir index and dir indices for all subdirs
            self.assertEqual(page.TYPE, "rst")
            self.assertCountEqual(page.tags, [tag_example, tag_another])
            self.assertEqual(
                page.to_dict(),
                {
                    "src": {
                        "abspath": os.path.join(
                            mocksite.site.content_root, "index.rst"
                        ),
                        "relpath": "index.rst",
                    },
                    "site_path": "",
                    "build_path": "index.html",
                    "meta": {
                        "author": "Test User",
                        "copyright": "© 2016 Test User",
                        "date": "2016-04-16 10:23:00+02:00",
                        "draft": False,
                        "syndicated": True,
                        "syndication_date": "2016-04-16 10:23:00+02:00",
                        "description": "line1\nline2\nline3",
                        "tags": [
                            "CategoryPage(taxonomies/tags/example)",
                            "CategoryPage(taxonomies/tags/another_tag)",
                        ],
                        "template": "page.html",
                        "template_copyright": "compiled:None",
                        "title": "Example blog post in reStructuredText",
                        "indexed": True,
                        "nav": [],
                        "site_name": "Test site",
                        "site_url": "https://www.example.org",
                        "related": {},
                    },
                    "type": "rst",
                },
            )



from __future__ import annotations

import os
from unittest import TestCase

from staticsite.cmd.build import Builder
from staticsite.site import Path

from . import utils as test_utils


class TestTutorial(test_utils.SiteTestMixin, TestCase):
    site_name = "tutorial"
    site_settings = {"SITE_AUTHOR": "Test User"}

    @test_utils.assert_no_logs()
    def test_render_paths(self):
        self.assertBuilt("index.md", "", "index.html", sample="Welcome to my new blog")


class BuiltExampleSite(test_utils.ExampleSite):
    def populate_workdir(self):
        super().populate_workdir()
        # Build the site inside the workdir
        site = self.create_site()
        site.settings.OUTPUT = os.path.join(self.root, "built_site")
        overrides = {"st_mtime": self.mock_file_mtime}
        with test_utils.mock_file_stat(overrides):
            site.load()
        builder = Builder(site)
        builder.write()


class TestBuiltTutorial(test_utils.SiteTestMixin, TestCase):
    site_name = "tutorial"
    site_settings = {"SITE_AUTHOR": "Test User"}
    site_cls = BuiltExampleSite

    @test_utils.assert_no_logs()
    def test_built_marker(self):
        built_marker = os.path.join(self.mocksite.root, "built_site", ".staticsite")
        self.assertTrue(os.path.exists(built_marker))
        with open(built_marker) as fd:
            self.assertEqual(fd.read(), "---\nskip: yes\n")

    @test_utils.assert_no_logs()
    def test_render_paths(self):
        self.assertBuilt("index.md", "", "index.html", sample="Welcome to my new blog")
        self.assertIsNone(self.site.root.lookup_page(Path.from_string("/built_site/")))



from __future__ import annotations

import io
from collections.abc import Callable
from typing import IO, Any, cast

#
# This is an unnecessarily complex wrapper to cope efficiently with YAML
# parsing in python.
#
# The current situation looks like:
#
#  * pyyaml is reasonably fast
#  * ruamel.yaml is unreasonably slow
#
# However, pyyaml doesn't support sort_keys until version 5.1, while
# ruamel.yaml can do unsorted dumping with no specific version requriements.
#
# So we mix and match, trying to use PyYAML for loading, and ruamel for
# dumping.
#

Load = Callable[[IO[str]], Any]
Loads = Callable[[str], Any]
Dump = Callable[[Any, IO[str]], None]
Dumps = Callable[[Any], str]

load_ruamel: Load | None
loads_ruamel: Loads | None
dump_ruamel: Dump | None
dumps_ruamel: Dumps | None

load_pyyaml: Load | None
loads_pyyaml: Loads | None
dump_pyyaml: Dump | None
dumps_pyyaml: Dumps | None

load: Load
loads: Loads
dump: Dump
dumps: Dumps

try:
    import ruamel.yaml

    yaml_loader = ruamel.yaml.YAML(typ="safe", pure=True)

    def loads_ruamel(string: str) -> Any:
        return yaml_loader.load(string)

    def load_ruamel(file: IO[str]) -> Any:
        return yaml_loader.load(file)

    # Hack to do unsorted serialization with ruamel
    yaml_dumper = ruamel.yaml.YAML(typ="rt", pure=True)
    yaml_dumper.allow_unicode = True
    yaml_dumper.default_flow_style = False
    # TODO: yaml does not have explicit_start typed correctly
    yaml_dumper.explicit_start = True  # type: ignore

    def dumps_ruamel(data: Any) -> str:
        # YAML dump with unsorted keys
        with io.StringIO() as fd:
            yaml_dumper.dump(data, fd)
            return fd.getvalue()

    def dump_ruamel(data: Any, file: IO[str]) -> None:
        yaml_dumper.dump(data, file)

except ModuleNotFoundError:
    load_ruamel = None
    loads_ruamel = None
    dump_ruamel = None
    dumps_ruamel = None

try:
    import yaml

    def loads_pyyaml(string: str) -> Any:
        return yaml.load(string, Loader=yaml.CLoader)

    def load_pyyaml(file: IO[str]) -> Any:
        return yaml.load(file, Loader=yaml.CLoader)

    def dumps_pyyaml(data: Any) -> str:
        # From pyyaml 5.1, one can add sort_keys=False
        # Before that version, it seems impossible to do unsorted serialization
        # with pyyaml
        # https://stackoverflow.com/questions/16782112/can-pyyaml-dump-dict-items-in-non-alphabetical-order
        return cast(
            str,
            yaml.dump(
                data,
                stream=None,
                default_flow_style=False,
                allow_unicode=True,
                explicit_start=True,
                Dumper=yaml.CDumper,
            ),
        )

    def dump_pyyaml(data: Any, file: IO[str]) -> None:
        yaml.dump(
            data,
            file,
            default_flow_style=False,
            allow_unicode=True,
            explicit_start=True,
            Dumper=yaml.CDumper,
        )

except ModuleNotFoundError:
    load_pyyaml = None
    loads_pyyaml = None
    dump_pyyaml = None
    dumps_pyyaml = None


# Prefer pyyaml for loading, because it's significantly faster
if load_pyyaml:
    load = load_pyyaml
elif load_ruamel:
    load = load_ruamel
else:
    raise RuntimeError("Neither PyYAML nor ruamel.YAML are installed")

if loads_pyyaml:
    loads = loads_pyyaml
elif loads_ruamel:
    loads = loads_ruamel
else:
    raise RuntimeError("Neither PyYAML nor ruamel.YAML are installed")


# Prefer ruamel for dumping, because it supports unsorted keys, that are useful
# when rendering archetypes
if dump_ruamel:
    dump = dump_ruamel
elif dump_pyyaml:
    dump = dump_pyyaml
else:
    raise RuntimeError("Neither PyYAML nor ruamel.YAML are installed")

if dumps_ruamel:
    dumps = dumps_ruamel
elif dumps_pyyaml:
    dumps = dumps_pyyaml
else:
    raise RuntimeError("Neither PyYAML nor ruamel.YAML are installed")



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .site import Site

log = logging.getLogger("archetypes")


class Archetype:
    def __init__(self, archetypes: Archetypes, relpath: str):
        self.archetypes = archetypes
        self.site = archetypes.site
        self.relpath = relpath

    def render(self, **kw: Any) -> tuple[dict[str, Any], str]:
        """
        Render the archetype with the given context information.

        Returns the metadata of the rendered page, and the rendered page
        contents.
        """
        # By default, render the archetype with jinja2
        abspath = os.path.join(self.archetypes.root, self.relpath)
        with open(abspath) as fd:
            template = self.site.theme.jinja2.from_string(fd.read())
        rendered = template.render(**kw)
        return {}, rendered


class Archetypes:
    def __init__(self, site: Site, root: str):
        self.site = site

        # Root directory where archetypes are found
        self.root = root

    def find(self, name: str) -> Archetype | None:
        """
        Read the archetypes directory and return the archetype that matches the given name.

        Returns None if nothing matches.
        """
        for root, dnames, fnames in os.walk(self.root, followlinks=True):
            for f in fnames:
                if f.startswith("."):
                    continue
                relpath = os.path.relpath(os.path.join(root, f), self.root)
                for feature in self.site.features.ordered():
                    a = feature.try_load_archetype(self, relpath, name)
                    if a is not None:
                        return a
        return None



from __future__ import annotations

import argparse
import contextlib
import locale
import logging
import os
import shutil
import time
from collections import Counter
from collections.abc import Generator
from typing import TYPE_CHECKING, Any

from .. import render, utils
from ..page import ChangeExtent
from ..site import Path
from .command import Fail, SiteCommand, register

if TYPE_CHECKING:
    from ..node import Node
    from ..page import Page
    from ..site import Site

log = logging.getLogger("build")


@register
class Build(SiteCommand):
    "build the site into the web/ directory of the project"

    @classmethod
    def add_subparser(
        cls, subparsers: argparse._SubParsersAction[Any]
    ) -> argparse.ArgumentParser:
        parser = super().add_subparser(subparsers)
        parser.add_argument(
            "--type", action="store", help="render only pages of this type"
        )
        parser.add_argument(
            "--path", action="store", help="render only pages under this path"
        )
        parser.add_argument(
            "--fail-fast",
            action="store_true",
            help="fail the first time a page gives an error in rendering",
        )
        parser.add_argument(
            "-f", "--full", action="store_true", help="always do a full rebuild"
        )
        return parser

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)
        self.site: Site

    def run(self) -> int | None:
        self.site = self.load_site()
        self.builder = Builder(
            self.site,
            type_filter=self.args.type,
            path_filter=self.args.path,
            fail_fast=self.args.fail_fast,
            full=self.args.full,
        )
        self.builder.write()
        if self.builder.has_errors:
            return 1
        return None


class RenderStats:
    """
    Statistics collected during rendering
    """

    def __init__(self) -> None:
        self.sums: dict[str, int] = Counter()
        self.counts: dict[str, int] = Counter()

    @contextlib.contextmanager
    def collect(self, page: Page) -> Generator[None, None, None]:
        start = time.perf_counter_ns()
        yield
        end = time.perf_counter_ns()
        self.sums[page.TYPE] += end - start
        self.counts[page.TYPE] += 1


class RenderDirectory:
    """
    A directory where contents are being rendered
    """

    def __init__(self, root: str, relpath: str, dir_fd: int):
        self.root = root
        self.relpath = relpath
        self.dir_fd = dir_fd

        # Scan directory contents
        self.old_dirs: dict[str, os.stat_result] = {}
        self.old_files: dict[str, os.stat_result] = {}
        with os.scandir(dir_fd) as entries:
            for de in entries:
                if de.is_dir():
                    self.old_dirs[de.name] = de.stat()
                else:
                    self.old_files[de.name] = de.stat()

    @classmethod
    @contextlib.contextmanager
    def open(cls, root: str) -> Generator[RenderDirectory, None, None]:
        """
        Start rendering in the given output root directory
        """
        os.makedirs(root, exist_ok=True)
        with utils.open_dir_fd(root) as dir_fd:
            yield cls(root, "", dir_fd)

    @contextlib.contextmanager
    def subdir(self, name: str) -> Generator[RenderDirectory, None, None]:
        subpath = os.path.join(self.relpath, name)
        with utils.open_dir_fd(name, dir_fd=self.dir_fd) as subdir_fd:
            yield RenderDirectory(self.root, subpath, subdir_fd)

    def prepare_subdir(self, name: str) -> os.stat_result | None:
        """
        Prepare for rendering a subdirectory.

        Return a File object corresponding to an existing directory entry
        """
        if name in self.old_dirs:
            # Directory already existed: reuse it
            return self.old_dirs.pop(name)
        elif name in self.old_files:
            # There was a file at this location: delete it
            self.old_files.pop(name, None)
            os.unlink(name, dir_fd=self.dir_fd)
            os.mkdir(name, dir_fd=self.dir_fd)
            return None
        else:
            # There was nothing: create the directory
            os.mkdir(name, dir_fd=self.dir_fd)
            return None

    def prepare_file(self, name: str) -> os.stat_result | None:
        """
        Prepare for rendering a file

        Return a File object corresponding to an existing directory entry
        """
        if name in self.old_dirs:
            # There is a directory instead: remove it
            self.old_dirs.pop(name, None)
            # FIXME: from Python 3.11, rmtree supports dir_fd
            shutil.rmtree(os.path.join(self.root, self.relpath, name))
            return None
        elif name in self.old_files:
            # There is a file at this location: it will be overwritten
            return self.old_files.pop(name)
        else:
            return None

    def cleanup_leftovers(self) -> None:
        """
        Remove previously existing files and directories that have disappeared
        in the current version of the site
        """
        for name in self.old_dirs:
            # FIXME: from Python 3.11, rmtree supports dir_fd
            shutil.rmtree(os.path.join(self.root, self.relpath, name))
        for name in self.old_files:
            os.unlink(name, dir_fd=self.dir_fd)


class Builder:
    def __init__(
        self,
        site: Site,
        type_filter: str | None = None,
        path_filter: str | None = None,
        fail_fast: bool = False,
        full: bool = True,
    ):
        self.site = site
        self.type_filter = type_filter
        self.path_filter = path_filter
        if site.settings.PROJECT_ROOT is None:
            raise Fail("PROJECT_ROOT is not set")
        if site.settings.OUTPUT is None:
            raise Fail(
                "No output directory configured:"
                " please use --output or set OUTPUT in settings.py or .staticsite.py"
            )
        self.build_root = os.path.join(site.settings.PROJECT_ROOT, site.settings.OUTPUT)
        # Logs which pages have been rendered and to which path
        self.build_log: dict[str, Page] = {}
        self.fail_fast = fail_fast
        self.full = full
        self.has_errors = False
        self.built_marker = render.RenderedString(
            """---
skip: yes
"""
        )

    def write(self) -> None:
        """
        Generate output
        """
        # Set locale for rendering
        try:
            lname = self.site.settings.LANGUAGES[0]["locale"]
            locale.setlocale(locale.LC_ALL, lname)
        except locale.Error as e:
            log.warning("Cannot set locale to %s: %s", lname, e)

        with utils.timings("Built site in %fs"):
            # cpu_count = os.cpu_count()
            # if cpu_count > 1:
            #     self.write_multi_process(cpu_count)
            # else:
            #     self.write_single_process()

            # I tried trivial parallelisation with child processes but the benefits
            # do not seem significant:
            #
            #     real  0m5.111s
            #     user  0m8.096s
            #     sys   0m0.644s
            #
            # compared with:
            #
            #     real  0m6.251s
            #     user  0m5.760s
            #     sys   0m0.468s
            self.write_single_process()

        with utils.timings("Saved build state in %fs"):
            if self.has_errors:
                # Output directory is partially build, a further build cannot rely on it
                self.site.clear_footprints()
            else:
                self.site.save_footprints()

        #     self.build_cache.put("git_hash", self.new_hexsha)
        #     self.build_cache.put("git_dirty", sorted(self.files_changed_in_workdir))
        #     # build_cache.put("git_dirty", repo.head.commit.hexsha)

    #     def write_multi_process(self, child_count):
    #         log.info("Generating pages using %d child processes", child_count)
    #
    #         pages = list(self.site.structure.pages.values())
    #
    #         # From http://code.activestate.com/recipes/576785-partition-an-iterable-into-n-lists/
    #         chunks = [pages[i::child_count] for i in range(child_count)]
    #
    #         print(len(pages))
    #         for c in chunks:
    #             print(len(c))
    #
    #         import sys
    #         pids = set()
    #         for chunk in chunks:
    #             pid = os.fork()
    #             if pid == 0:
    #                 self.write_pages(chunk)
    #                 sys.exit(0)
    #             else:
    #                 pids.add(pid)
    #
    #         while pids:
    #             (pid, status) = os.wait()
    #             pids.discard(pid)

    def write_single_process(self) -> None:
        root: Node = self.site.root
        if self.path_filter is not None:
            if (node := root.lookup_node(Path.from_string(self.path_filter))) is None:
                raise Fail(
                    f"path filter {self.path_filter} does not match a path in the site"
                )
            root = node
        stats = RenderStats()
        os.makedirs(self.build_root, exist_ok=True)

        with RenderDirectory.open(self.build_root) as render_dir:
            # Write built marker
            old_file = render_dir.prepare_file(".staticsite")
            self.built_marker.write(
                name=".staticsite", dir_fd=render_dir.dir_fd, old=old_file
            )

            # Write rendered contents
            self.write_subtree(root, render_dir, stats=stats)
        for type in sorted(stats.sums.keys()):
            log.info(
                "%s: %d in %.3fs (%.1f per minute)",
                type,
                stats.counts[type],
                stats.sums[type] / 1_000_000_000,
                stats.counts[type] / stats.sums[type] * 60 * 1_000_000_000,
            )

    def write_subtree(
        self, node: Node, render_dir: RenderDirectory, stats: RenderStats
    ) -> None:
        """
        Recursively render the given node in the given render directory
        """
        log.debug("write_subtree relpath:%s node:%r", render_dir.relpath, node)
        # If this is the build node for a page, render it
        for name, page in node.build_pages.items():
            if self.type_filter and page.TYPE != self.type_filter:
                continue
            if (old_file := render_dir.prepare_file(name)) is not None:
                # TODO: simple minded so far
                if not self.full and page.change_extent == ChangeExtent.UNCHANGED:
                    continue
            with stats.collect(page):
                try:
                    rendered = page.render()
                except Exception:
                    if self.fail_fast:
                        raise
                    log.error(
                        "%s:%s page failed to render",
                        render_dir.relpath,
                        name,
                        exc_info=True,
                    )
                    self.has_errors = True
                else:
                    # log.debug("write_subtree relpath:%s render %s %s", render_dir.relpath, page.TYPE, name)
                    rendered.write(name=name, dir_fd=render_dir.dir_fd, old=old_file)
                self.build_log[os.path.join(render_dir.relpath, name)] = page

        for name, sub in node.sub.items():
            # Subdir
            # log.debug("write_subtree relpath:%s render subdir %s", render_dir.relpath, name)
            render_dir.prepare_subdir(name)
            with render_dir.subdir(name) as subdir:
                self.write_subtree(sub, subdir, stats)

        render_dir.cleanup_leftovers()

    def output_abspath(self, relpath: str) -> str:
        abspath = os.path.join(self.build_root, relpath)
        os.makedirs(os.path.dirname(abspath), exist_ok=True)
        return abspath



from __future__ import annotations

import argparse
import logging
from typing import TYPE_CHECKING, Any

from .command import Fail, SiteCommand, register

if TYPE_CHECKING:
    from ..site import Site

log = logging.getLogger("dump")


@register
class Dump(SiteCommand):
    "Dump information about a site"

    @classmethod
    def add_subparser(
        cls, subparsers: argparse._SubParsersAction[Any]
    ) -> argparse.ArgumentParser:
        parser = super().add_subparser(subparsers)
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument(
            "--fstree",
            action="store_true",
            help="dump information about the scanned directory trees",
        )
        group.add_argument(
            "--nodes",
            action="store_true",
            help="dump information about the built site layout",
        )
        group.add_argument(
            "--changes",
            action="store_true",
            help="dump information about pages that changed",
        )
        group.add_argument(
            "--reference-documentation",
            action="store",
            nargs="?",
            const="doc/reference",
            metavar="destdir",
            help="regenerate reference documentation",
        )
        return parser

    def dump_fstree(self, site: Site) -> None:
        for name, fstree in site.fstrees.items():
            print(f"# {name}")
            fstree.print()

    def dump_nodes(self, site: Site) -> None:
        site.root.print()

    def dump_changes(self, site: Site) -> None:
        for page in site.iter_pages():
            change = page.change_extent
            if change > 0:
                print(f"{page!r} {change}")

    def run(self) -> None:
        site = self.load_site()
        if self.args.fstree:
            self.dump_fstree(site)
        elif self.args.nodes:
            self.dump_nodes(site)
        elif self.args.changes:
            self.dump_changes(site)
        elif self.args.reference_documentation:
            from ..autodoc import Autodoc

            autodoc = Autodoc(site, self.args.reference_documentation)
            autodoc.generate()
        else:
            raise Fail("I don't know what to do")



from unittest import TestCase

import toml

from staticsite.utils import front_matter

toml_valid = """+++
date = "Tue, 24 Aug 2010 13:13:38 +0000"
tags = [ "valid" ]
bool = true
+++
"""

toml_invalid_bool = """+++
date = "Tue, 24 Aug 2010 13:13:38 +0000"
tags = [ "bug" ]
bool = True
+++
"""


class TestToml(TestCase):
    def test_valid(self):
        self.maxDiff = None
        fmt, data = front_matter.read_string(toml_valid)
        self.assertEqual(fmt, "toml")
        self.assertEqual(
            data,
            {
                "date": "Tue, 24 Aug 2010 13:13:38 +0000",
                "tags": ["valid"],
                "bool": True,
            },
        )

    def test_invalid_bool(self):
        with self.assertRaises(toml.decoder.TomlDecodeError):
            fmt, data = front_matter.read_string(toml_invalid_bool)



from __future__ import annotations

import argparse
import logging
import shlex
import subprocess
from typing import TYPE_CHECKING, Any, cast

from staticsite.page_filter import PageFilter

from ..page import SourcePage
from .command import Fail, SiteCommand, register

if TYPE_CHECKING:
    from staticsite.features.taxonomy import TaxonomyFeature


log = logging.getLogger("edit")


@register
class Edit(SiteCommand):
    "edit an existing page. Bare keywords match titles and file names, '+tag' matches tags"

    MENU_SIZE = 5

    def select_page_menu(self, pages: list[SourcePage]) -> SourcePage | None:
        """
        Prompt the user to select a page from `pages`.

        Returns None of the user chooses to give up insteaad.
        """
        first = 0

        while True:
            pagination = len(pages) > self.MENU_SIZE

            if pagination:
                print(
                    "{count} pages found, showing {first}-{last}:".format(
                        first=first + 1,
                        last=min(first + self.MENU_SIZE, len(pages)),
                        count=len(pages),
                    )
                )
                prompt = (
                    "Please choose which page to edit"
                    " (number, or p for previous page, n for next page, q to quit): "
                )
            else:
                print(f"{len(pages)} matching pages found:")
                prompt = "Please choose which page to edit (number, or q to quit): "
            print()

            for idx, page in enumerate(
                pages[first : first + self.MENU_SIZE], first + 1
            ):
                print(
                    "{idx}: {date} {relpath} {title}".format(
                        idx=idx,
                        date=page.meta["date"].strftime("%Y-%m-%d"),
                        relpath=page.src.relpath,
                        title=page.meta.get("title", "{no title}"),
                    )
                )
            print()

            selection = input(prompt).strip().lower()
            if selection == "q":
                return None
            elif pagination and selection == "p":
                first = max(0, first - self.MENU_SIZE)
                continue
            elif pagination and selection == "n" or selection == "":
                first = min(
                    first + self.MENU_SIZE,
                    self.MENU_SIZE * (len(pages) // self.MENU_SIZE),
                )
                continue

            if selection.isdigit():
                idx = int(selection)
                if idx < 1 or idx > len(pages):
                    print(f"{idx} does not match any page")
                else:
                    return pages[idx - 1]

    def run(self) -> int | None:
        site = self.load_site()

        filter_args: dict[str, Any] = {
            "sort": "-date",
        }

        # Filter by taxonomy
        taxonomies: TaxonomyFeature = cast("TaxonomyFeature", site.features["taxonomy"])
        args = []
        for arg in self.args.match:
            if arg.startswith("+"):
                filter_args.setdefault("tags", []).append(arg[1:])
            else:
                for name in taxonomies.taxonomies.keys():
                    if arg.startswith(f"{name}:"):
                        filter_args.setdefault("tags", []).append(arg[len(name) + 1 :])
                        break
                else:
                    args.append(arg)

        def match_page(page: SourcePage) -> bool:
            for m in args:
                if (
                    m.lower() not in page.meta.get("title", "").lower()
                    and m.lower() not in page.src.relpath.lower()
                ):
                    return False
            return True

        # Build a list of all findable pages, present on disk, sorted with the newest first
        f = PageFilter(site, **filter_args)
        pages = [
            page
            for page in f.filter()
            if isinstance(page, SourcePage) and match_page(page)
        ]

        if len(pages) == 0:
            raise Fail(
                "No page found matching {}".format(
                    " ".join(shlex.quote(x) for x in self.args.match)
                )
            )

        page: SourcePage | None
        if len(pages) == 1:
            page = pages[0]
        else:
            page = self.select_page_menu(pages)

        if page is None:
            return None

        abspath = page.src.abspath

        if not self.args.noedit:
            settings_dict = site.settings.as_dict()
            cmd = [
                x.format(name=abspath, **settings_dict)
                for x in site.settings.EDIT_COMMAND
            ]
            try:
                subprocess.check_call(cmd)
            except subprocess.CalledProcessError as e:
                log.warning(
                    "Editor command %s exited with error %d",
                    " ".join(shlex.quote(x) for x in cmd),
                    e.returncode,
                )
                return e.returncode
        print(abspath)
        return None

    @classmethod
    def add_subparser(
        cls, subparsers: argparse._SubParsersAction[Any]
    ) -> argparse.ArgumentParser:
        parser = super().add_subparser(subparsers)
        parser.add_argument(
            "match", nargs="*", help="keywords used to look for the page to edit"
        )
        parser.add_argument(
            "-n",
            "--noedit",
            action="store_true",
            help="do not run an editor, only output the file name of the new post",
        )
        return parser



from __future__ import annotations

import datetime
import logging
import os
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Optional, TypeVar, cast

import jinja2

from staticsite import fields
from staticsite.feature import Feature, PageTrackingMixin, TrackedField
from staticsite.page import (
    AutoPage,
    ChangeExtent,
    Page,
    PageNotFoundError,
    TemplatePage,
)

if TYPE_CHECKING:
    from staticsite.features import rst

log = logging.getLogger("syndication")

P = TypeVar("P", bound="Page")


class Syndication:
    """
    Syndication information for a group of pages
    """

    def __init__(
        self,
        index: SyndicationPageMixin | None,
        *,
        title: str | None = None,
        description: str | None = None,
        template_title: str | None = None,
        template_description: str | None = None,
        add_to: bool | dict[str, Any] | None = True,
        archive: bool | dict[str, Any] = True,
    ):
        # Default metadata for syndication pages
        self.default_meta: dict[str, Any] = {}
        if title is not None:
            self.default_meta["title"] = title
        if description is not None:
            self.default_meta["description"] = description
        if template_title is not None:
            self.default_meta["template_title"] = template_title
        if template_description is not None:
            self.default_meta["template_description"] = template_description

        # Page that defined the syndication
        self.index: SyndicationPageMixin | None = index

        # Pages that get syndicated
        self.pages: list[SyndicationPageMixin] = []

        # Pages to which we add feed information
        #
        # If True: defaults to self.pages.
        #
        # If a string: expanded as a page filter
        self.add_to = add_to

        # Describe if and how to build an archive for this syndication
        self.archive: dict[str, Any] | None
        if archive is False:
            self.archive = None
        elif archive is True:
            self.archive = {}
        elif isinstance(archive, dict):
            self.archive = archive
        else:
            raise ValueError(
                f"{archive!r} is not a valid value for a syndication.archive field"
            )

        # RSS feed page
        self.rss_page: RSSPage | None = None

        # Atom feed page
        self.atom_page: AtomPage | None = None

        # Archive page
        self.archive_page: ArchivePage | None = None

    def to_dict(self) -> dict[str, Any]:
        return {
            "index": self.index,
            "pages": self.pages,
            "add_to": self.add_to,
            "archive": self.archive,
            "rss_page": self.rss_page,
            "atom_page": self.atom_page,
            "archive_page": self.archive_page,
        }

    def generate(self) -> None:
        """
        Create autogenerated pages
        """
        # Generate RSS and Atom feeds
        self.make_feeds()

        # Make archive pages
        self.make_archive()

    def make_feeds(self) -> None:
        """
        Build RSS and Atom feeds
        """
        if self.index is None:
            raise RuntimeError("Syndication.index field was not set")
        if (
            "title" not in self.default_meta
            and "template_title" not in self.default_meta
        ):
            self.default_meta["title"] = self.index.title

        if (
            "description" not in self.default_meta
            and "template_description" not in self.default_meta
            and self.index.description
        ):
            self.default_meta["description"] = self.index.description

        page_name, ext = os.path.splitext(self.index.dst)

        kwargs = self.default_meta
        kwargs["pages"] = self.pages
        kwargs["index"] = self.index

        # RSS feed
        rss_page = self.index.node.create_auto_page_as_file(
            created_from=self.index,
            page_cls=RSSPage,
            dst=f"{page_name}.{RSSPage.TYPE}",
            **kwargs,
        )
        self.rss_page = rss_page
        if rss_page is not None:
            log.debug("%s: adding syndication page for %s", rss_page, self.index)
            self.index.add_related("rss_feed", rss_page)

        # print(f"  rss_page {rss_page.meta=!r}")

        # Atom feed
        atom_page = self.index.node.create_auto_page_as_file(
            created_from=self.index,
            page_cls=AtomPage,
            dst=f"{page_name}.{AtomPage.TYPE}",
            **kwargs,
        )
        self.atom_page = atom_page
        if atom_page is not None:
            log.debug("%s: adding syndication page for %s", atom_page, self.index)
            self.index.add_related("atom_feed", atom_page)

    def make_archive(self) -> None:
        """
        Build archive pages
        """
        if self.archive is None:
            return
        if self.index is None:
            raise RuntimeError("Syndication.index field was not set")

        self.archive["pages"] = self.pages
        self.archive["index"] = self.index
        self.archive_page = self.index.node.create_auto_page_as_path(
            created_from=self.index,
            page_cls=ArchivePage,
            name="archive",
            **self.archive,
        )
        if self.archive_page is not None:
            if self.rss_page is not None:
                self.archive_page.add_related("rss_feed", self.rss_page)
            if self.atom_page is not None:
                self.archive_page.add_related("atom_feed", self.atom_page)

    def crossreference(self) -> None:
        """
        Cross-correlate generated pages
        """
        # Build the final list of the pages covered by the syndication
        self.fill_pages()

        # Add feed pages according to syndication.add_to
        self.process_add_to()

    def fill_pages(self) -> None:
        """
        Given a list of pages to potentially syndicate, filter them by their
        syndicated header, and sort by syndication date.
        """
        if self.index is None:
            raise RuntimeError("Syndication.index field was not set")
        draft_mode = self.index.site.settings.DRAFT_MODE
        for page in self.index.pages or ():
            spage = cast(SyndicationPageMixin, page)
            if not spage.syndicated:
                continue
            if not draft_mode and spage.syndication_date > spage.site.generation_time:
                continue
            self.pages.append(spage)
        self.pages.sort(key=lambda p: p.syndication_date, reverse=True)

        # Compute dates for generated pages
        if self.pages:
            max_date = max(p.date for p in self.pages)
        else:
            max_date = self.index.site.generation_time

        if self.rss_page is not None:
            self.rss_page.date = max_date
        if self.atom_page is not None:
            self.atom_page.date = max_date
        if self.archive_page is not None:
            self.archive_page.date = max_date

    def process_add_to(self) -> None:
        """
        Add a link to the syndication to the pages listed in add_to
        """
        if self.index is None:
            raise RuntimeError("Syndication.index field was not set")
        add_to = self.add_to
        if add_to is False or add_to in ("no", "false", 0):
            pass
        elif add_to is True:
            for dest in self.pages:
                if self.rss_page is not None:
                    dest.add_related("rss_feed", self.rss_page)
                if self.atom_page is not None:
                    dest.add_related("atom_feed", self.atom_page)
        elif isinstance(add_to, dict):
            for pdest in self.index.find_pages(**add_to):
                if self.rss_page is not None:
                    pdest.add_related("rss_feed", self.rss_page)
                if self.atom_page is not None:
                    pdest.add_related("atom_feed", self.atom_page)

    @classmethod
    def clean_value(
        self, page: SyndicationPageMixin | None, value: Any
    ) -> Syndication | None:
        """
        Instantiate a Syndication object from a Page field
        """
        if value is True:
            return Syndication(page)
        elif value is False:
            return None
        elif isinstance(value, Syndication):
            value.index = page
            return value
        elif isinstance(value, str):
            if value.lower() in ("yes", "true", "1"):
                return Syndication(page)
            else:
                return None
        elif isinstance(value, dict):
            return Syndication(page, **value)
        else:
            raise ValueError(
                f"{value!r} for `syndication` needs to be True or a dictionary of values"
            )


class SyndicatedPageError(Exception):
    pass


class SyndicationField(TrackedField["SyndicationPageMixin", Optional[Syndication]]):
    """
    Defines syndication for the contents of this page.

    It is a structure which can contain normal metadata, plus:

    * `add_to`: chooses which pages will include a link to the RSS/Atom feeds.
      By default, the link is added to all syndicated pages. If this is `False`,
      then no feed is added to pages. If it is a dictionary, it selects pages in
      the site, similar to the `site_pages` function in [templates](templates.md).
      See [Selecting pages](page-filter.md) for details.
    * `archive`: if not false, an archive link will be generated next to the
      page. It can be set of a dictionary of metadata to be used as defaults for
      the generated archive page. It defaults to True.

    Any other metadata found in the structure are used when generating pages for
    the RSS/Atom feeds, so you can use `title`, `template_title`, `description`,
    and so on, to personalize the feeds.

    The pages that go in the feed are those listed in
    [`page.meta.pages`](doc/reference/pages.md), keeping into account the
    [`syndicated` and `syndication_date` page metadata](doc/reference/metadata.md).

    When rendering RSS/Atom feed pages, `page.meta.pages` is replaced with the list
    of syndicated pages, sorted with the most recent first.

    Setting `syndication` to true turns on syndication with all defaults,
    equivalent to:

    ```yaml
    syndication:
      add_to: yes
      archive:
        template: archive.html
    ```
    """

    tracked_by = "syndication"

    def _clean(self, page: SyndicationPageMixin, value: Any) -> Syndication | None:
        return Syndication.clean_value(page, value)


class SyndicationDateField(fields.Date[Page]):
    """
    Syndication date for this page.

    This is the date that will appear in RSS and Atom feeds, and the page will not
    be syndicated before this date.

    Use this field to publish a date today but make it appear in
    syndication only at some point in the future.

    If a page is syndicated and `syndication_date` is missing, it defaults to `date`.
    """

    def __get__(self, page: Page, type: type[Page] | None = None) -> datetime.datetime:
        if (date := page.__dict__.get(self.name)) is None:
            date = page.date
            page.__dict__[self.name] = date
        return cast(datetime.datetime, date)


class SyndicatedField(fields.Bool[Page]):
    """
    Set to true if the page can be included in a syndication, else to false.

    If not set, it defaults to the value of `indexed`.
    """

    def __get__(self, page: Page, type: type[Page] | None = None) -> bool:
        if (cur := page.__dict__.get(self.name)) is None:
            value = cast(bool, page.indexed)
            page.__dict__[self.name] = value
            return value
        else:
            return cast(bool, cur)


class SyndicationPageMixin(Page):
    syndication = SyndicationField(structure=True)
    syndicated = SyndicatedField(
        doc="""
        Set to true if the page can be included in a syndication, else to false.

        If not set, it defaults to the value of `indexed`.
    """
    )
    syndication_date = SyndicationDateField()
    # rss_page = fields.Field(doc="""
    #     Page with the RSS feed with posts for the syndication the page is in
    # """)
    # atom_page = fields.Field(doc="""
    #     Page with the Atom feed with posts for the syndication the page is in
    # """)
    # archive = fields.Field(doc="""
    #     Page with an archive of all posts for the syndication the page is in
    # """)


class SyndicationIndexField(fields.Field[P, SyndicationPageMixin]):
    """
    Field containing a Taxonomy object
    """

    def _clean(self, page: P, value: Any) -> SyndicationPageMixin:
        if isinstance(value, SyndicationPageMixin):
            return value
        else:
            raise TypeError(
                f"invalid value of type {type(value)} for {page!r}.{self.name}:"
                " expecting SyndicationPageMixin"
            )


def _get_syndicated_pages(
    page: SyndicationPageMixin, limit: int | None = None
) -> Sequence[Page]:
    """
    Get the sorted list of syndicated pages for a page
    """
    if (syndication := page.syndication) is not None:
        # syndication.pages is already sorted
        return syndication.pages[:limit]

    if (pages := page.pages) is None:
        raise SyndicatedPageError(
            f"page {page!r} has no `syndication.pages` or `pages` in metadata"
        )
    return pages.arrange("-syndication_date", limit=limit)


class SyndicationFeature(PageTrackingMixin[SyndicationPageMixin], Feature):
    r"""
    Build syndication feeds for groups of pages.

    The Syndication feature implements generation of a [RSS](https://en.wikipedia.org/wiki/RSS)
    and [Atom](https://en.wikipedia.org/wiki/Atom_(Web_standard\)) feed for a page
    or group of pages.

    Add a `syndication` metadata to a page to declare it as the index for a group
    of page. RSS and Atom feeds will appear next to the page, containing pages
    selected in the `syndication` metadata.

    One page is used to define the syndication, using "syndication_*" tags.

    Use a data page without type to define a contentless syndication page

    ## Syndication metadata

    Example front page for a blog page:

    ```yaml
    ---
    pages: blog/*
    syndication: yes
    template: blog.html
    ---
    # My blog
    ```

    See [metadata documentation](metadata.md) for a reference on the `syndication`
    field.

    ## Syndication of taxonomies

    Each category page in each taxonomy automatically defines a syndication
    metadata equivalent to this:

    ```yaml
    syndication:
      add_to: no
    ```

    This would automatically generates RSS and Atom feeds with all pages in that
    category, but those feed links will only be added to the category page itself.

    You can use the `syndication` metadata in your taxonomy categories to customize
    titles and description in your categories feeds, like with any other page.


    ## Templates

    See the example templates for working `syndication.rss` and `syndication.atom`
    templates, that are used by the generated RSS and Atom pages.

    ### `syndicated_pages(page=None, limit=None)`

    Templates can use the `syndicated_pages` function to list syndicated pages for
    a page, sorted with the most recently syndicated first.

    `page` can be a page, a path to a page, or omitted, in which case the current
    page is used. `page` can also be a list of pages, which will be sorted by
    syndiaction date and sampled.

    `limit` is the number of pages to return, or, if omitted, all the pages are
    returned.
    """
    RUN_AFTER = ["rst"]

    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)
        cast("rst.RestructuredText", self.site.features["rst"]).yaml_tags.add(
            "syndication"
        )
        self.syndications: list[Syndication] = []
        self.j2_globals["syndicated_pages"] = self.jinja2_syndicated_pages

    def get_page_bases(self, page_cls: type[Page]) -> Sequence[type[Page]]:
        return (SyndicationPageMixin,)

    def get_used_page_types(self) -> list[type[Page]]:
        return [RSSPage, AtomPage, ArchivePage]

    @jinja2.pass_context
    def jinja2_syndicated_pages(
        self,
        context: jinja2.runtime.Context,
        what: str | SyndicationPageMixin | jinja2.Undefined | None = None,
        limit: int | None = None,
    ) -> Sequence[Page]:
        """
        Get the list of pages to be syndicated
        """
        try:
            if what is None or isinstance(what, jinja2.Undefined):
                if (page := context.get("page")) is None:
                    raise SyndicatedPageError(
                        "syndicated_pages called without a page argument, but page is not set in context"
                    )
                return _get_syndicated_pages(
                    cast(SyndicationPageMixin, page), limit=limit
                )
            elif isinstance(what, str):
                src = context.get("page")
                if src is None:
                    raise SyndicatedPageError(
                        "syndicated_pages called without a page argument, but page is not set in context"
                    )

                try:
                    page = src.resolve_path(what)
                except PageNotFoundError as e:
                    log.warning("%s: %s", context.name, e)
                    return []

                return _get_syndicated_pages(page, limit=limit)
            elif isinstance(what, Page):
                return _get_syndicated_pages(what, limit=limit)
            else:
                raise SyndicatedPageError(
                    f"syndicated_pages called with a page argument of unsupported type {what.__class__.__name__}"
                )
        except SyndicatedPageError as e:
            log.warning("%s: %s", context.name, e)
            return []

    def generate(self) -> None:
        # Build syndications from pages with a 'syndication' metadata
        for page in self.tracked_pages:
            # The syndication header is the base for the feed pages's metadata,
            # and is added as 'syndication' to all the pages that get the feed
            # links
            if (syndication := page.syndication) is None:
                continue

            syndication.generate()

            self.syndications.append(syndication)

    def crossreference(self) -> None:
        for syndication in self.syndications:
            syndication.crossreference()


class SyndicationPage(TemplatePage, AutoPage):
    """
    Base class for syndication pages
    """

    index = SyndicationIndexField["SyndicationPage"](
        doc="""
        Page that defined the syndication for this feed
    """
    )

    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)
        if self.pages:
            self.date = max(p.date for p in self.pages)
        else:
            self.date = self.site.generation_time

    def _compute_change_extent(self) -> ChangeExtent:
        if self.created_from is None:
            raise AssertionError(f"SyndicationPage {self!r} found without created_from")
        return self.created_from._compute_change_extent()


class RSSPage(SyndicationPage):
    """
    A RSS syndication page

    RSS and Atom pages have these extra properties:

    * `page.meta.template` defaults to `syndication.rss` or `syndication.atom`
      instead of `page.html`
    * `page.meta.date` is the most recent date of the pages in the feed
    * `page.meta.index` is the page defining the syndication
    * `page.meta.pages` is a list of all the pages included in the syndication
    """

    TYPE = "rss"
    TEMPLATE = "syndication.rss"


class AtomPage(SyndicationPage):
    """
    An Atom syndication page

    RSS and Atom pages have these extra properties:

    * `page.meta.template` defaults to `syndication.rss` or `syndication.atom`
      instead of `page.html`
    * `page.meta.date` is the most recent date of the pages in the feed
    * `page.meta.index` is the page defining the syndication
    * `page.meta.pages` is a list of all the pages included in the syndication
    """

    TYPE = "atom"
    TEMPLATE = "syndication.atom"


class ArchivePage(TemplatePage, AutoPage):
    """
    An archive page is automatically created for each syndication.

    * `page.meta.pages`: the pages in the archive
    * `page.created_from`: the page for which the archive page was created
    """

    index = SyndicationIndexField["SyndicationPage"](
        doc="""
        Page that defined the syndication for this archive
    """
    )

    TYPE = "archive"
    TEMPLATE = "archive.html"

    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)

        if self.created_from is None:
            raise AssertionError(f"ArchivePage {self!r} has empty created_from")
        self.pages = self.created_from.pages

        self.created_from.add_related("archive", self)

    def _compute_change_extent(self) -> ChangeExtent:
        if self.created_from is None:
            raise AssertionError(f"ArchivePage {self!r} found without created_from")
        return self.created_from._compute_change_extent()


FEATURES = {
    "syndication": SyndicationFeature,
}



from __future__ import annotations

import json
import logging
import os
import shlex
import subprocess
from typing import TYPE_CHECKING, Any

import PIL
import PIL.Image

if TYPE_CHECKING:
    from staticsite.cache import Cache
    from staticsite.file import File

log = logging.getLogger("utils.images")


class ImageScanner:
    def __init__(self, cache: Cache):
        self.cache = cache

    def scan(self, src: File, mimetype: str) -> dict[str, Any]:
        key = f"{src.abspath}:{src.stat.st_mtime:.3f}"
        meta: dict[str, Any] | None = self.cache.get(key)
        if meta is None:
            meta = self.read_meta(src.abspath, mimetype)
            self.cache.put(key, meta)
        return meta

    def scan_file(self, pathname: str) -> dict[str, Any]:
        import mimetypes

        mimetypes.init()
        base, ext = os.path.splitext(pathname)
        mimetype = mimetypes.types_map.get(ext)
        if mimetype is None:
            return {}
        return self.read_meta(pathname, mimetype)

    def read_meta(self, pathname: str, mimetype: str) -> dict[str, Any]:
        # We can take our time here, since results are cached

        if mimetype == "image/svg+xml":
            return {}

        with PIL.Image.open(pathname) as img:
            meta = {
                "width": img.width,
                "height": img.height,
                "title": "",
            }

            if img.format not in ("XPM",):
                meta.update(self.read_meta_exiftool(pathname))

        return meta

    def read_meta_exiftool(self, pathname: str) -> dict[str, Any]:
        meta: dict[str, Any] = {}

        # It is important to use abspath here, as exiftool does not support the
        # usual -- convention to deal with files starting with a dash. With abspath
        # at least the file name will start with a /
        res = subprocess.run(
            ["exiftool", "-json", "-n", "-c", "%f", pathname], capture_output=True
        )
        if res.returncode != 0:
            log.warning(
                "%s: exiftool failed with code %d: %s",
                pathname,
                res.returncode,
                res.stderr.strip(),
            )
            return meta

        info = json.loads(res.stdout)[0]

        description = info.get("ImageDescription")
        if description is not None:
            meta["title"] = description

        artist = info.get("Artist")
        if artist is not None:
            meta["author"] = artist

        orientation = info.get("Orientation")
        if orientation is not None:
            # https://www.impulseadventure.com/photo/exif-orientation.html
            meta["image_orientation"] = int(orientation)

        copyright = info.get("CopyrightNotice")
        if copyright is not None:
            meta["copyright"] = copyright

        lat = info.get("GPSLatitude")
        if lat is not None:
            meta["lat"] = float(lat)

        lon = info.get("GPSLongitude")
        if lon is not None:
            meta["lon"] = float(lon)

        # "DateTime": "2017:05:09 21:27:42",
        # "GPSLatitudeRef": "North",
        # "GPSAltitude": "92 m",
        # "GPSTimeStamp": "19:27:43",
        # "GPSDateStamp": "2017:05:09",
        # "GPSDateTime": "2017:05:09 19:27:43Z",

        return meta

    def edit_meta_exiftool(
        self, pathname: str, changed: dict[str, Any], removed: list[str]
    ) -> bool:
        exif_args: list[str] = []

        if "title" in changed:
            exif_args.append(f"-ImageDescription={changed['title']}")

        if "author" in changed:
            exif_args.append(f"-Artist={changed['author']}")

        if "image_orientation" in changed:
            exif_args.append(f"-Orientation={changed['image_orientation']}")

        if "copyright" in changed:
            # See https://libre-software.net/edit-metadata-exiftool/
            exif_args.append(f"-rights={changed['copyright']}")
            exif_args.append(f"-CopyrightNotice={changed['copyright']}")

        # lat = info.get("GPSLatitude")
        # if lat is not None:
        #     print("EXIF LAT", lat)

        # lon = info.get("GPSLongitude")
        # if lon is not None:
        #     print("EXIF LON", lon)

        for name in removed:
            if name == "title":
                exif_args.append("-ImageDescription=")
            elif name == "author":
                exif_args.append("-Artist=")
            elif name == "image_orientation":
                exif_args.append("-Orientation=")
            elif name == "copyright":
                exif_args.append("-rights=")
                exif_args.append("-CopyrightNotice=")

        cmd = [
            "exiftool",
            "-c",
            "%f",
            "-overwrite_original",
            "-quiet",
            pathname,
        ] + exif_args
        res = subprocess.run(cmd)
        if res.returncode != 0:
            log.warning(
                "%s: %s failed with code %d: %s",
                pathname,
                " ".join(shlex.quote(x) for x in cmd),
                res.returncode,
            )
            return False

        return True
